import mock
import netaddr

from quark.db import api as db_api
from quark.tests.functional.mysql.base import MySqlBaseFunctionalTest
from quark.tools import null_routes


class TestNullRoutes(MySqlBaseFunctionalTest):
    def setUp(self):
        super(TestNullRoutes, self).setUp()
        self.cidr = "192.168.10.0/24"
        self.sub_cidr = "192.168.10.1/32"

    def test_get_subnets_cidr_set(self):
        network = db_api.network_create(self.context)
        db_api.subnet_create(
            self.context,
            network=network, cidr=self.cidr)
        self.context.session.flush()
        ipset = null_routes.get_subnets_cidr_set(self.context, [network.id])
        self.assertEqual(ipset, netaddr.IPSet(netaddr.IPNetwork(self.cidr)))

    @mock.patch("requests.get")
    def test_get_null_routes_addresses(self, get_method):
        url, region = "TEST_URL", "TEST_REGION"
        ipset = netaddr.IPSet(netaddr.IPNetwork(self.cidr))
        payload = [{
            "status": "1",
            "note": None,
            "updated": None,
            "name": None,
            "status_name": None,
            "region.id": region,
            "ip": None,
            "idql": None,
            "discovered": None,
            "netmask": None,
            "tag": None,
            "conf": None,
            "cidr": self.sub_cidr,
            "id": None,
            "switch.hostname": None,
        }]
        body = [{
            "paginate": {
                "total_count": len(payload),
                "total_count_display": None,
                "total_pages": None,
                "author_comment": None,
                "per_page": None,
                "page": None
            },
            "request": None,
            "payload": payload,
            "response": None
        }]
        get_method.return_value.json.return_value = body
        addresses = null_routes.get_null_routes_addresses(url, region, ipset)
        get_method.assert_called_once_with(url, verify=False)
        self.assertEqual(addresses,
                         netaddr.IPSet(netaddr.IPNetwork(self.sub_cidr)))

    def test_delete_locks_has_lock(self):
        network = db_api.network_create(self.context)
        address_model = db_api.ip_address_create(
            self.context,
            address=netaddr.IPAddress("192.168.10.1"),
            network=network)
        db_api.lock_holder_create(
            self.context, address_model,
            name=null_routes.LOCK_NAME, type="ip_address")
        self.context.session.flush()

        null_routes.delete_locks(self.context, [network.id], [])
        self.context.session.refresh(address_model)
        self.assertIsNone(address_model.lock_id)

    def test_delete_locks_doesnt_have_lock(self):
        network = db_api.network_create(self.context)
        address_model = db_api.ip_address_create(
            self.context,
            address=netaddr.IPAddress("192.168.10.1"),
            network=network)
        db_api.lock_holder_create(
            self.context, address_model,
            name="not-null-routes", type="ip_address")
        self.context.session.flush()

        null_routes.delete_locks(self.context, [network.id], [])
        self.context.session.refresh(address_model)
        self.assertIsNotNone(address_model.lock_id)

    def test_create_locks_address_doesnt_exist(self):
        network = db_api.network_create(self.context)
        subnet = db_api.subnet_create(
            self.context,
            network=network,
            cidr=self.cidr,
            ip_version=4)
        self.context.session.flush()

        addresses = netaddr.IPSet(netaddr.IPNetwork(self.sub_cidr))
        null_routes.create_locks(self.context, [network.id], addresses)
        address = db_api.ip_address_find(
            self.context, subnet_id=subnet.id, scope=db_api.ONE)
        self.assertIsNotNone(address)
        self.assertIsNotNone(address.lock_id)

    def test_create_locks_address_exists(self):
        network = db_api.network_create(self.context)
        address_model = db_api.ip_address_create(
            self.context,
            address=netaddr.IPAddress("192.168.10.1"),
            network=network)
        self.context.session.flush()

        addresses = netaddr.IPSet(netaddr.IPNetwork(self.sub_cidr))
        null_routes.create_locks(self.context, [network.id], addresses)
        self.context.session.refresh(address_model)
        self.assertIsNotNone(address_model.lock_id)

    def test_create_locks_lock_holder_exists(self):
        network = db_api.network_create(self.context)
        address_model = db_api.ip_address_create(
            self.context,
            address=netaddr.IPAddress("192.168.10.1"),
            network=network)
        db_api.lock_holder_create(
            self.context, address_model,
            name=null_routes.LOCK_NAME, type="ip_address")
        self.context.session.flush()

        addresses = netaddr.IPSet(netaddr.IPNetwork(self.sub_cidr))
        null_routes.create_locks(self.context, [network.id], addresses)

        lock_holders = db_api.lock_holder_find(
            self.context,
            lock_id=address_model.lock_id,
            name=null_routes.LOCK_NAME,
            scope=db_api.ALL)
        self.assertEqual(len(lock_holders), 1)



import contextlib
import datetime
import os
import tempfile

from alembic import command as alembic_command
from alembic import config as alembic_config
import mock
import netaddr
import sqlalchemy as sa
from sqlalchemy import create_engine
from sqlalchemy import pool
from sqlalchemy.sql import column
from sqlalchemy.sql import select
from sqlalchemy.sql import table

from quark.db.custom_types import INET
import quark.db.migration
from quark.tests import test_base


class BaseMigrationTest(test_base.TestBase):
    def setUp(self):
        self.config = alembic_config.Config(
            os.path.join(quark.db.migration.__path__[0], 'alembic.ini'))
        self.config.set_main_option('script_location',
                                    'quark.db.migration:alembic')
        self.config.set_main_option("quiet_mode", "True")
        self.fileno, self.filepath = tempfile.mkstemp()
        secret_cfg = mock.MagicMock()
        secret_cfg.database.connection = "sqlite:///" + self.filepath
        self.config.neutron_config = secret_cfg

        self.engine = create_engine(
            self.config.neutron_config.database.connection,
            poolclass=pool.NullPool)
        self.connection = self.engine.connect()

    def tearDown(self):
        self.connection.close()
        os.unlink(self.filepath)


class Test2748e48cee3a(BaseMigrationTest):
    def setUp(self):
        super(Test2748e48cee3a, self).setUp()
        alembic_command.upgrade(self.config, '1284c81cf727')
        self.ip_policy_cidrs = table(
            'quark_ip_policy_cidrs',
            column('id', sa.String(length=36)),
            column('created_at', sa.DateTime()),
            column('ip_policy_id', sa.String(length=36)),
            column('cidr', sa.String(length=64)))
        self.subnets = table(
            'quark_subnets',
            column('id', sa.String(length=36)),
            column('_cidr', sa.String(length=64)),
            column('ip_policy_id', sa.String(length=36)))

    def test_upgrade_no_ip_policy_cidr(self):
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", _cidr="192.168.10.0/24", ip_policy_id=None))

        alembic_command.upgrade(self.config, '2748e48cee3a')
        results = self.connection.execute(
            select([self.ip_policy_cidrs])).fetchall()
        self.assertEqual(len(results), 0)

    def test_upgrade_ip_policy_cidr_inside(self):
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", _cidr="192.168.10.0/24", ip_policy_id="111"))
        dt = datetime.datetime(1970, 1, 1)
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="222", created_at=dt,
                 ip_policy_id="111", cidr="192.168.10.0/32"))

        alembic_command.upgrade(self.config, '2748e48cee3a')
        results = self.connection.execute(
            select([self.ip_policy_cidrs])).fetchall()
        self.assertEqual(len(results), 1)
        result = results[0]
        self.assertEqual(result["id"], "222")
        self.assertEqual(result["created_at"], dt)
        self.assertEqual(result["ip_policy_id"], "111")
        self.assertEqual(result["cidr"], "192.168.10.0/32")

    def test_upgrade_ip_policy_cidr_overlaps(self):
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", _cidr="192.168.10.0/24", ip_policy_id="111"))
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="222", created_at=datetime.date(1970, 1, 1),
                 ip_policy_id="111", cidr="192.168.10.0/16"))

        with contextlib.nested(
            mock.patch("oslo_utils.uuidutils"),
            mock.patch("oslo_utils.timeutils")
        ) as (uuid, tu):
            tu.utcnow.return_value = datetime.datetime(2004, 2, 14)
            uuid.generate_uuid.return_value = "foo"
            alembic_command.upgrade(self.config, '2748e48cee3a')
            results = self.connection.execute(
                select([self.ip_policy_cidrs])).fetchall()
            self.assertEqual(len(results), 1)
            result = results[0]
            self.assertEqual(result["id"], uuid.generate_uuid.return_value)
            self.assertEqual(result["created_at"], tu.utcnow.return_value)
            self.assertEqual(result["ip_policy_id"], "111")
            self.assertEqual(result["cidr"], "192.168.10.0/24")

    def test_upgrade_ip_policy_cidr_overlaps_v6(self):
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", _cidr="fd00::/8", ip_policy_id="111"))
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="222", created_at=datetime.date(1970, 1, 1),
                 ip_policy_id="111", cidr="fd00::/7"))

        with contextlib.nested(
            mock.patch("oslo_utils.uuidutils"),
            mock.patch("oslo_utils.timeutils")
        ) as (uuid, tu):
            tu.utcnow.return_value = datetime.datetime(2004, 2, 14)
            uuid.generate_uuid.return_value = "foo"
            alembic_command.upgrade(self.config, '2748e48cee3a')
            results = self.connection.execute(
                select([self.ip_policy_cidrs])).fetchall()
            self.assertEqual(len(results), 1)
            result = results[0]
            self.assertEqual(result["id"], uuid.generate_uuid.return_value)
            self.assertEqual(result["created_at"], tu.utcnow.return_value)
            self.assertEqual(result["ip_policy_id"], "111")
            self.assertEqual(result["cidr"], "fd00::/8")

    def test_upgrade_ip_policy_cidr_outside(self):
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", _cidr="192.168.10.0/24", ip_policy_id="111"))
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="222", created_at=datetime.date(1970, 1, 1),
                 ip_policy_id="111", cidr="0.0.0.0/24"))

        alembic_command.upgrade(self.config, '2748e48cee3a')
        results = self.connection.execute(
            select([self.ip_policy_cidrs])).fetchall()
        self.assertEqual(len(results), 0)

    def test_upgrade_bulk(self):
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", _cidr="192.168.10.0/24", ip_policy_id=None),
            dict(id="001", _cidr="192.168.10.0/24", ip_policy_id="111"),
            dict(id="002", _cidr="192.168.10.0/24", ip_policy_id="112"),
            dict(id="003", _cidr="192.168.10.0/24", ip_policy_id="113"))
        dt = datetime.datetime(1970, 1, 1)
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="221", created_at=dt, ip_policy_id="111",
                 cidr="192.168.10.0/32"),
            dict(id="222", created_at=dt, ip_policy_id="112",
                 cidr="192.168.10.0/16"),
            dict(id="223", created_at=dt, ip_policy_id="113",
                 cidr="0.0.0.0/24"))

        with contextlib.nested(
            mock.patch("oslo_utils.uuidutils"),
            mock.patch("oslo_utils.timeutils")
        ) as (uuid, tu):
            tu.utcnow.return_value = datetime.datetime(2004, 2, 14)
            uuid.generate_uuid.return_value = "foo"
            alembic_command.upgrade(self.config, '2748e48cee3a')
            results = self.connection.execute(
                select([self.ip_policy_cidrs])).fetchall()
            self.assertEqual(len(results), 2)
            result = results[0] if results[0]["id"] == "foo" else results[1]
            self.assertEqual(result["id"], uuid.generate_uuid.return_value)
            self.assertEqual(result["created_at"], tu.utcnow.return_value)
            self.assertEqual(result["ip_policy_id"], "112")
            self.assertEqual(result["cidr"], "192.168.10.0/24")
            result = results[0] if results[0]["id"] != "foo" else results[1]
            self.assertEqual(result["id"], "221")
            self.assertEqual(result["created_at"], dt)
            self.assertEqual(result["ip_policy_id"], "111")
            self.assertEqual(result["cidr"], "192.168.10.0/32")

    def test_upgrade_multiple_ip_policy_cidrs(self):
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", _cidr="192.168.10.0/24", ip_policy_id="111"))
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="221", created_at=datetime.date(1970, 1, 1),
                 ip_policy_id="111", cidr="0.0.0.0/24"),
            dict(id="222", created_at=datetime.date(1970, 1, 1),
                 ip_policy_id="111", cidr="192.168.10.255/32"),
            dict(id="223", created_at=datetime.date(1970, 1, 1),
                 ip_policy_id="111", cidr="192.168.10.0/23"))

        with contextlib.nested(
            mock.patch("oslo_utils.uuidutils"),
            mock.patch("oslo_utils.timeutils")
        ) as (uuid, tu):
            tu.utcnow.return_value = datetime.datetime(2004, 2, 14)
            uuid.generate_uuid.return_value = "foo"
            alembic_command.upgrade(self.config, '2748e48cee3a')
            results = self.connection.execute(
                select([self.ip_policy_cidrs])).fetchall()
            self.assertEqual(len(results), 1)
            result = results[0]
            self.assertEqual(result["id"], uuid.generate_uuid.return_value)
            self.assertEqual(result["created_at"], tu.utcnow.return_value)
            self.assertEqual(result["ip_policy_id"], "111")
            self.assertEqual(result["cidr"], "192.168.10.0/24")

    def test_downgrade(self):
        alembic_command.upgrade(self.config, '2748e48cee3a')
        with self.assertRaises(NotImplementedError):
            alembic_command.downgrade(self.config, '1284c81cf727')


class Test45a07fac3d38(BaseMigrationTest):
    def setUp(self):
        super(Test45a07fac3d38, self).setUp()
        alembic_command.upgrade(self.config, '2748e48cee3a')
        self.ip_policy_cidrs = table(
            'quark_ip_policy_cidrs',
            column('id', sa.String(length=36)),
            column('created_at', sa.DateTime()),
            column('ip_policy_id', sa.String(length=36)),
            column('cidr', sa.String(length=64)))
        self.subnets = table(
            'quark_subnets',
            column('id', sa.String(length=36)),
            column('_cidr', sa.String(length=64)),
            column('ip_policy_id', sa.String(length=36)))

    def test_upgrade_no_subnets_no_ip_policy_cidrs(self):
        alembic_command.upgrade(self.config, '45a07fac3d38')
        results = self.connection.execute(
            select([self.ip_policy_cidrs])).fetchall()
        self.assertEqual(len(results), 0)

    def test_upgrade_with_subnets_no_ip_policy(self):
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", _cidr="192.168.10.0/24", ip_policy_id=None))
        alembic_command.upgrade(self.config, '45a07fac3d38')
        results = self.connection.execute(
            select([self.ip_policy_cidrs])).fetchall()
        self.assertEqual(len(results), 0)

    def test_upgrade_with_subnets_no_ip_policy_cidrs(self):
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", _cidr="192.168.10.0/24", ip_policy_id="111"))
        alembic_command.upgrade(self.config, '45a07fac3d38')
        results = self.connection.execute(
            select([self.ip_policy_cidrs])).fetchall()
        self.assertEqual(len(results), 2)
        default_cidrs = ["192.168.10.0/32", "192.168.10.255/32"]
        self.assertIn(results[0]["cidr"], default_cidrs)
        self.assertIn(results[1]["cidr"], default_cidrs)
        self.assertNotEqual(results[0]["cidr"], results[1]["cidr"])

    def test_upgrade_with_subnets_non_default_ip_policy_cidrs(self):
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", _cidr="192.168.10.0/24", ip_policy_id="111"))
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="222", created_at=datetime.date(1970, 1, 1),
                 ip_policy_id="111", cidr="192.168.10.13/32"))
        with contextlib.nested(
            mock.patch("oslo_utils.uuidutils"),
            mock.patch("oslo_utils.timeutils")
        ) as (uuid, tu):
            uuid.generate_uuid.side_effect = (1, 2, 3)
            tu.utcnow.return_value = datetime.datetime(1970, 1, 1)
            alembic_command.upgrade(self.config, '45a07fac3d38')
            results = self.connection.execute(
                select([self.ip_policy_cidrs])).fetchall()
            self.assertEqual(len(results), 3)
            default_cidrs = ["192.168.10.0/32", "192.168.10.255/32",
                             "192.168.10.13/32"]
            for result in results:
                self.assertIn(result["cidr"], default_cidrs)
                self.assertGreaterEqual(int(result["id"]), 1)
                self.assertLessEqual(int(result["id"]), 3)
                self.assertEqual(result["created_at"], tu.utcnow.return_value)
            self.assertNotEqual(results[0]["cidr"], results[1]["cidr"])
            self.assertNotEqual(results[0]["cidr"], results[2]["cidr"])
            self.assertNotEqual(results[1]["cidr"], results[2]["cidr"])

    def test_upgrade_with_subnets_non_default_ip_policy_cidrs_v6(self):
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", _cidr="fd00::/64", ip_policy_id="111"))
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="222", created_at=datetime.date(1970, 1, 1),
                 ip_policy_id="111", cidr="fd00::3/128"))
        with contextlib.nested(
            mock.patch("oslo_utils.uuidutils"),
            mock.patch("oslo_utils.timeutils")
        ) as (uuid, tu):
            uuid.generate_uuid.side_effect = (1, 2, 3)
            tu.utcnow.return_value = datetime.datetime(1970, 1, 1)
            alembic_command.upgrade(self.config, '45a07fac3d38')
            results = self.connection.execute(
                select([self.ip_policy_cidrs])).fetchall()
            self.assertEqual(len(results), 3)
            default_cidrs = ["fd00::/128", "fd00::3/128",
                             "fd00::ffff:ffff:ffff:ffff/128"]
            for result in results:
                self.assertIn(result["cidr"], default_cidrs)
                self.assertGreaterEqual(int(result["id"]), 1)
                self.assertLessEqual(int(result["id"]), 3)
                self.assertEqual(result["created_at"], tu.utcnow.return_value)
            self.assertNotEqual(results[0]["cidr"], results[1]["cidr"])
            self.assertNotEqual(results[0]["cidr"], results[2]["cidr"])
            self.assertNotEqual(results[1]["cidr"], results[2]["cidr"])

    def test_upgrade_with_subnets_default_ip_policy_cidrs(self):
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", _cidr="192.168.10.0/24", ip_policy_id="111"))
        dt = datetime.datetime(1970, 1, 1)
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="222", created_at=dt,
                 ip_policy_id="111", cidr="192.168.10.0/32"),
            dict(id="223", created_at=dt,
                 ip_policy_id="111", cidr="192.168.10.255/32"))
        alembic_command.upgrade(self.config, '45a07fac3d38')
        results = self.connection.execute(
            select([self.ip_policy_cidrs])).fetchall()
        self.assertEqual(len(results), 2)
        default_cidrs = ["192.168.10.0/32", "192.168.10.255/32"]
        self.assertIn(results[0]["cidr"], default_cidrs)
        self.assertIn(results[1]["cidr"], default_cidrs)
        self.assertTrue(results[0]["id"] == "222" or results[0]["id"] == "223")
        self.assertTrue(results[1]["id"] == "222" or results[1]["id"] == "223")
        self.assertEqual(results[0]["created_at"], dt)
        self.assertEqual(results[1]["created_at"], dt)

    def test_upgrade_bulk(self):
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", _cidr="192.168.10.0/24", ip_policy_id=None),
            dict(id="001", _cidr="192.168.10.0/24", ip_policy_id="111"),
            dict(id="002", _cidr="192.168.10.0/24", ip_policy_id="112"),
            dict(id="003", _cidr="192.168.10.0/24", ip_policy_id="113"))
        dt = datetime.datetime(1970, 1, 1)
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="221", created_at=dt,
                 ip_policy_id="112", cidr="192.168.10.13/32"),
            dict(id="222", created_at=dt,
                 ip_policy_id="113", cidr="192.168.10.0/32"),
            dict(id="223", created_at=dt,
                 ip_policy_id="113", cidr="192.168.10.255/32"))

        alembic_command.upgrade(self.config, '45a07fac3d38')

        results = self.connection.execute(
            select([self.ip_policy_cidrs]).where(
                self.ip_policy_cidrs.c.ip_policy_id == None)).fetchall()  # noqa
        self.assertEqual(len(results), 0)

        results = self.connection.execute(
            select([self.ip_policy_cidrs]).where(
                self.ip_policy_cidrs.c.ip_policy_id == "111")).fetchall()
        self.assertEqual(len(results), 2)
        default_cidrs = ["192.168.10.0/32", "192.168.10.255/32"]
        self.assertIn(results[0]["cidr"], default_cidrs)
        self.assertIn(results[1]["cidr"], default_cidrs)
        self.assertNotEqual(results[0]["cidr"], results[1]["cidr"])

        results = self.connection.execute(
            select([self.ip_policy_cidrs]).where(
                self.ip_policy_cidrs.c.ip_policy_id == "112")).fetchall()
        self.assertEqual(len(results), 3)
        default_cidrs = ["192.168.10.0/32", "192.168.10.255/32",
                         "192.168.10.13/32"]
        for result in results:
            self.assertIn(result["cidr"], default_cidrs)
        self.assertNotEqual(results[0]["cidr"], results[1]["cidr"])
        self.assertNotEqual(results[0]["cidr"], results[2]["cidr"])
        self.assertNotEqual(results[1]["cidr"], results[2]["cidr"])

        results = self.connection.execute(
            select([self.ip_policy_cidrs]).where(
                self.ip_policy_cidrs.c.ip_policy_id == "113")).fetchall()
        self.assertEqual(len(results), 2)
        default_cidrs = ["192.168.10.0/32", "192.168.10.255/32"]
        self.assertIn(results[0]["cidr"], default_cidrs)
        self.assertIn(results[1]["cidr"], default_cidrs)
        self.assertTrue(results[0]["id"] == "222" or results[0]["id"] == "223")
        self.assertTrue(results[1]["id"] == "222" or results[1]["id"] == "223")
        self.assertEqual(results[0]["created_at"], dt)
        self.assertEqual(results[1]["created_at"], dt)

    def test_downgrade(self):
        alembic_command.upgrade(self.config, '45a07fac3d38')
        with self.assertRaises(NotImplementedError):
            alembic_command.downgrade(self.config, '2748e48cee3a')


class Test552b213c2b8c(BaseMigrationTest):
    def setUp(self):
        super(Test552b213c2b8c, self).setUp()
        alembic_command.upgrade(self.config, '45a07fac3d38')
        self.ip_policy = table(
            'quark_ip_policy',
            column('id', sa.String(length=36)),
            column('tenant_id', sa.String(length=255)),
            column('created_at', sa.DateTime()))
        self.ip_policy_cidrs = table(
            'quark_ip_policy_cidrs',
            column('id', sa.String(length=36)),
            column('created_at', sa.DateTime()),
            column('ip_policy_id', sa.String(length=36)),
            column('cidr', sa.String(length=64)))
        self.subnets = table(
            'quark_subnets',
            column('id', sa.String(length=36)),
            column('tenant_id', sa.String(length=255)),
            column('_cidr', sa.String(length=64)),
            column('ip_policy_id', sa.String(length=36)))

    def test_upgrade_no_subnets(self):
        alembic_command.upgrade(self.config, '552b213c2b8c')
        results = self.connection.execute(
            select([self.ip_policy_cidrs])).fetchall()
        self.assertEqual(len(results), 0)

    def test_upgrade_subnets_with_ip_policy(self):
        dt = datetime.datetime(1970, 1, 1)
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", tenant_id="foo", _cidr="192.168.10.0/24",
                 ip_policy_id="111"))
        self.connection.execute(
            self.ip_policy.insert(),
            dict(id="111", tenant_id="foo", created_at=dt))
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="221", created_at=dt,
                 ip_policy_id="111", cidr="192.168.10.13/32"))
        alembic_command.upgrade(self.config, '552b213c2b8c')
        results = self.connection.execute(
            select([self.ip_policy])).fetchall()
        self.assertEqual(len(results), 1)
        result = results[0]
        self.assertEqual(result["id"], "111")
        self.assertEqual(result["tenant_id"], "foo")
        self.assertEqual(result["created_at"], dt)
        results = self.connection.execute(
            select([self.ip_policy_cidrs])).fetchall()
        self.assertEqual(len(results), 1)
        result = results[0]
        self.assertEqual(result["id"], "221")
        self.assertEqual(result["created_at"], dt)
        self.assertEqual(result["ip_policy_id"], "111")
        self.assertEqual(result["cidr"], "192.168.10.13/32")
        results = self.connection.execute(
            select([self.subnets])).fetchall()
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]["ip_policy_id"], "111")

    def test_upgrade_subnets_no_ip_policy(self):
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", tenant_id="foo", _cidr="192.168.10.0/24",
                 ip_policy_id=None))
        with contextlib.nested(
            mock.patch("oslo_utils.uuidutils"),
            mock.patch("oslo_utils.timeutils")
        ) as (uuid, tu):
            dt = datetime.datetime(1970, 1, 1)
            tu.utcnow.return_value = dt
            uuid.generate_uuid.side_effect = ("666", "667", "668")
            alembic_command.upgrade(self.config, '552b213c2b8c')
        results = self.connection.execute(
            select([self.ip_policy])).fetchall()
        self.assertEqual(len(results), 1)
        result = results[0]
        self.assertEqual(result["id"], "666")
        self.assertEqual(result["tenant_id"], "foo")
        self.assertEqual(result["created_at"], dt)
        results = self.connection.execute(
            select([self.ip_policy_cidrs])).fetchall()
        self.assertEqual(len(results), 2)
        for result in results:
            self.assertIn(result["id"], ("667", "668"))
            self.assertEqual(result["created_at"], dt)
            self.assertEqual(result["ip_policy_id"], "666")
            self.assertIn(result["cidr"],
                          ("192.168.10.0/32", "192.168.10.255/32"))
        self.assertNotEqual(results[0]["cidr"], results[1]["cidr"])
        results = self.connection.execute(
            select([self.subnets])).fetchall()
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]["ip_policy_id"], "666")

    def test_upgrade_subnets_no_ip_policy_v6(self):
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", tenant_id="foo", _cidr="fd00::/64",
                 ip_policy_id=None))
        with contextlib.nested(
            mock.patch("oslo_utils.uuidutils"),
            mock.patch("oslo_utils.timeutils")
        ) as (uuid, tu):
            dt = datetime.datetime(1970, 1, 1)
            tu.utcnow.return_value = dt
            uuid.generate_uuid.side_effect = ("666", "667", "668")
            alembic_command.upgrade(self.config, '552b213c2b8c')
        results = self.connection.execute(
            select([self.ip_policy])).fetchall()
        self.assertEqual(len(results), 1)
        result = results[0]
        self.assertEqual(result["id"], "666")
        self.assertEqual(result["tenant_id"], "foo")
        self.assertEqual(result["created_at"], dt)
        results = self.connection.execute(
            select([self.ip_policy_cidrs])).fetchall()
        self.assertEqual(len(results), 2)
        for result in results:
            self.assertIn(result["id"], ("667", "668"))
            self.assertEqual(result["created_at"], dt)
            self.assertEqual(result["ip_policy_id"], "666")
            self.assertIn(result["cidr"],
                          ("fd00::/128",
                           "fd00::ffff:ffff:ffff:ffff/128"))
        self.assertNotEqual(results[0]["cidr"], results[1]["cidr"])
        results = self.connection.execute(
            select([self.subnets])).fetchall()
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]["ip_policy_id"], "666")

    def test_upgrade_bulk(self):
        dt = datetime.datetime(1970, 1, 1)
        self.connection.execute(
            self.subnets.insert(),
            dict(id="000", tenant_id="foo", _cidr="192.168.10.0/24",
                 ip_policy_id="111"),
            dict(id="001", tenant_id="foo", _cidr="192.168.10.0/24",
                 ip_policy_id=None),
            dict(id="002", tenant_id="foo", _cidr="fd00::/64",
                 ip_policy_id=None))
        self.connection.execute(
            self.ip_policy.insert(),
            dict(id="111", tenant_id="foo", created_at=dt))
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="221", created_at=dt,
                 ip_policy_id="111", cidr="192.168.10.13/32"))

        with contextlib.nested(
            mock.patch("oslo_utils.uuidutils"),
            mock.patch("oslo_utils.timeutils")
        ) as (uuid, tu):
            tu.utcnow.return_value = dt
            uuid.generate_uuid.side_effect = ("5", "6", "7", "8", "9", "10")
            alembic_command.upgrade(self.config, '552b213c2b8c')

        results = self.connection.execute(
            select([self.ip_policy]).where(
                self.ip_policy.c.id == "111")).fetchall()
        self.assertEqual(len(results), 1)
        result = results[0]
        self.assertEqual(result["id"], "111")
        self.assertEqual(result["tenant_id"], "foo")
        self.assertEqual(result["created_at"], dt)
        results = self.connection.execute(
            select([self.ip_policy_cidrs]).where(
                self.ip_policy_cidrs.c.ip_policy_id == "111")).fetchall()
        self.assertEqual(len(results), 1)
        result = results[0]
        self.assertEqual(result["id"], "221")
        self.assertEqual(result["created_at"], dt)
        self.assertEqual(result["ip_policy_id"], "111")
        self.assertEqual(result["cidr"], "192.168.10.13/32")
        results = self.connection.execute(
            select([self.subnets]).where(
                self.subnets.c.ip_policy_id == "111")).fetchall()
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]["ip_policy_id"], "111")

        results = self.connection.execute(
            select([self.ip_policy]).where(
                self.ip_policy.c.id != "111")).fetchall()
        self.assertEqual(len(results), 2)
        for result in results:
            self.assertIn(int(result["id"]), range(5, 11))
            self.assertEqual(result["tenant_id"], "foo")
            self.assertEqual(result["created_at"], dt)
        results = self.connection.execute(
            select([self.ip_policy_cidrs]).where(
                self.ip_policy_cidrs.c.ip_policy_id != "111")).fetchall()
        self.assertEqual(len(results), 4)
        for result in results:
            self.assertIn(int(result["id"]), range(5, 11))
            self.assertEqual(result["created_at"], dt)
            self.assertIn(int(result["ip_policy_id"]), range(5, 11))
            self.assertIn(result["cidr"], (
                "192.168.10.0/32", "192.168.10.255/32",
                "fd00::/128", "fd00::ffff:ffff:ffff:ffff/128"))
        results = self.connection.execute(
            select([self.subnets]).where(
                self.subnets.c.ip_policy_id != "111")).fetchall()
        self.assertEqual(len(results), 2)
        for subnet in results:
            self.assertIn(int(subnet["ip_policy_id"]), range(5, 11))

    def test_downgrade(self):
        alembic_command.upgrade(self.config, '552b213c2b8c')
        with self.assertRaises(NotImplementedError):
            alembic_command.downgrade(self.config, '45a07fac3d38')


class Test28e55acaf366(BaseMigrationTest):
    def setUp(self):
        super(Test28e55acaf366, self).setUp()
        alembic_command.upgrade(self.config, '3d22de205729')
        self.ip_policy = table('quark_ip_policy',
                               column('id', sa.String(length=36)),
                               column('size', INET()))
        self.ip_policy_cidrs = table(
            'quark_ip_policy_cidrs',
            column('id', sa.String(length=36)),
            column('ip_policy_id', sa.String(length=36)),
            column('cidr', sa.String(length=64)))

    def test_upgrade_none(self):
        alembic_command.upgrade(self.config, '28e55acaf366')
        results = self.connection.execute(select([
            self.ip_policy])).fetchall()
        self.assertEqual(len(results), 0)
        results = self.connection.execute(select([
            self.ip_policy_cidrs])).fetchall()
        self.assertEqual(len(results), 0)

    def test_upgrade_v4(self):
        self.connection.execute(
            self.ip_policy.insert(), dict(id="1", size=None))
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="2", ip_policy_id="1", cidr="192.168.10.13/32"),
            dict(id="3", ip_policy_id="1", cidr="192.168.10.16/31"))
        alembic_command.upgrade(self.config, '28e55acaf366')
        results = self.connection.execute(select([
            self.ip_policy])).fetchall()
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]["id"], "1")
        self.assertEqual(results[0]["size"], 3)

    def test_upgrade_v6(self):
        self.connection.execute(
            self.ip_policy.insert(), dict(id="1", size=None))
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="2", ip_policy_id="1", cidr="fd00::/64"))
        alembic_command.upgrade(self.config, '28e55acaf366')
        results = self.connection.execute(select([
            self.ip_policy])).fetchall()
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]["id"], "1")
        self.assertEqual(results[0]["size"], 2 ** 64)

    def test_upgrade_bulk(self):
        self.connection.execute(
            self.ip_policy.insert(),
            dict(id="1", size=None),
            dict(id="2", size=None))
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="2", ip_policy_id="1", cidr="192.168.10.13/32"),
            dict(id="3", ip_policy_id="1", cidr="192.168.10.16/31"),
            dict(id="4", ip_policy_id="2", cidr="fd00::/64"))
        alembic_command.upgrade(self.config, '28e55acaf366')
        results = self.connection.execute(select([
            self.ip_policy])).fetchall()
        self.assertEqual(len(results), 2)
        for result in results:
            self.assertIn(result["id"], ("1", "2"))
            if result["id"] == "1":
                self.assertEqual(result["size"], 3)
            elif result["id"] == "2":
                self.assertEqual(result["size"], 2 ** 64)

    def test_downgrade(self):
        alembic_command.upgrade(self.config, '28e55acaf366')
        with self.assertRaises(NotImplementedError):
            alembic_command.downgrade(self.config, '3d22de205729')


class Test1664300cb03a(BaseMigrationTest):
    def setUp(self):
        super(Test1664300cb03a, self).setUp()
        alembic_command.upgrade(self.config, '1acd075bd7e1')
        self.ip_policy_cidrs = table(
            'quark_ip_policy_cidrs',
            column('id', sa.String(length=36)),
            column('ip_policy_id', sa.String(length=36)),
            column('cidr', sa.String(length=64)),
            column('first_ip', INET()),
            column('last_ip', INET()))

    def test_upgrade_empty(self):
        alembic_command.upgrade(self.config, '1664300cb03a')
        results = self.connection.execute(select([
            self.ip_policy_cidrs])).fetchall()
        self.assertEqual(len(results), 0)

    def test_upgrade_ipv4(self):
        net = netaddr.IPNetwork("192.168.10.13/31")
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="1", ip_policy_id="1", cidr=str(net)))
        alembic_command.upgrade(self.config, '1664300cb03a')
        results = self.connection.execute(select([
            self.ip_policy_cidrs])).fetchall()
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]["id"], "1")
        self.assertEqual(results[0]["ip_policy_id"], "1")
        self.assertEqual(results[0]["cidr"], str(net))
        self.assertEqual(results[0]["first_ip"], net.ipv6().first)
        self.assertEqual(results[0]["last_ip"], net.ipv6().last)

    def test_upgrade_ipv6(self):
        net = netaddr.IPNetwork("fd00::/64")
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="1", ip_policy_id="1", cidr=str(net)))
        alembic_command.upgrade(self.config, '1664300cb03a')
        results = self.connection.execute(select([
            self.ip_policy_cidrs])).fetchall()
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]["id"], "1")
        self.assertEqual(results[0]["ip_policy_id"], "1")
        self.assertEqual(results[0]["cidr"], "fd00::/64")
        self.assertEqual(results[0]["first_ip"], net.first)
        self.assertEqual(results[0]["last_ip"], net.last)

    def test_upgrade_bulk(self):
        netv4 = netaddr.IPNetwork("192.168.10.13/31")
        netv6 = netaddr.IPNetwork("fd00::/64")
        self.connection.execute(
            self.ip_policy_cidrs.insert(),
            dict(id="1", ip_policy_id="1", cidr=str(netv4)),
            dict(id="2", ip_policy_id="2", cidr=str(netv6)))
        alembic_command.upgrade(self.config, '1664300cb03a')
        results = self.connection.execute(select([
            self.ip_policy_cidrs])).fetchall()
        self.assertEqual(len(results), 2)
        for result in results:
            self.assertIn(result["cidr"], (str(netv4), str(netv6)))
            if result["cidr"] == "192.168.10.13/31":
                self.assertEqual(result["first_ip"], netv4.ipv6().first)
                self.assertEqual(result["last_ip"], netv4.ipv6().last)
            else:
                self.assertEqual(result["first_ip"], netv6.first)
                self.assertEqual(result["last_ip"], netv6.last)

    def test_downgrade(self):
        alembic_command.upgrade(self.config, '1664300cb03a')
        with self.assertRaises(NotImplementedError):
            alembic_command.downgrade(self.config, '1acd075bd7e1')


class Test4fc07b41d45c(BaseMigrationTest):
    def _mock_inserts(self):
        mock_rows = [dict(id="1", _deallocated=False, address_type=None),
                     dict(id="2", _deallocated=True, address_type=None),
                     dict(id="3", _deallocated=None, address_type=None),
                     dict(id="4", _deallocated=True, address_type=None),
                     dict(id="5", _deallocated=False, address_type=None)]
        [insert.execute() for insert in
            [self.ip_addresses_table.insert().values(**row)
             for row in mock_rows]]

    def setUp(self):
        super(Test4fc07b41d45c, self).setUp()
        self.previous_revision = "42a3c8c0db75"
        self.current_revision = "4fc07b41d45c"
        self.metadata = sa.MetaData(bind=self.engine)
        # NOTE(thomasem): Create a quark_ip_addresses table that has an
        # identical schema as the revision before it for the columns this data
        # migration is concerned with.
        self.ip_addresses_table = sa.Table(
            'quark_ip_addresses', self.metadata,
            sa.Column('id', sa.String(length=36), primary_key=True),
            sa.Column('_deallocated', sa.Boolean()),
            sa.Column('address_type', sa.Enum('fixed', 'shared', 'floating'))
        )
        self.metadata.create_all()
        alembic_command.stamp(self.config, self.previous_revision)

    def test_upgrade(self):
        self._mock_inserts()
        alembic_command.upgrade(self.config, self.current_revision)
        results = self.connection.execute(
            select([self.ip_addresses_table]).order_by(
                self.ip_addresses_table.c.id)).fetchall()
        expected_results = [
            (u'1', False, u'fixed'),
            (u'2', True, None),
            (u'3', None, None),
            (u'4', True, None),
            (u'5', False, u'fixed')
        ]
        self.assertEqual(results, expected_results)

    def test_downgrade(self):
        self._mock_inserts()
        alembic_command.upgrade(self.config, self.current_revision)
        alembic_command.downgrade(self.config, self.previous_revision)
        results = self.connection.execute(
            select([self.ip_addresses_table]).order_by(
                self.ip_addresses_table.c.id)).fetchall()
        expected_results = [
            (u'1', False, None),
            (u'2', True, None),
            (u'3', None, None),
            (u'4', True, None),
            (u'5', False, None)
        ]
        self.assertEqual(results, expected_results)

    def test_upgrade_empty(self):
        alembic_command.upgrade(self.config, self.current_revision)
        results = self.connection.execute(
            select([self.ip_addresses_table]).order_by(
                self.ip_addresses_table.c.id)).fetchall()
        expected_results = []
        self.assertEqual(results, expected_results)



# Copyright 2012 New Dream Network, LLC (DreamHost)
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import six

from alembic import command as alembic_command
from alembic import config as alembic_config
from alembic import environment
from alembic import script as alembic_script
from alembic import util as alembic_util
from oslo_config import cfg

HEAD_FILENAME = 'HEAD'

_core_opts = [
    cfg.StrOpt('core_plugin',
               default='',
               help=_('Neutron plugin provider module')),
]

_quota_opts = [
    cfg.StrOpt('quota_driver',
               default='',
               help=_('Neutron quota driver class')),
]

_db_opts = [
    cfg.StrOpt('connection',
               deprecated_name='sql_connection',
               default='',
               secret=True,
               help=_('URL to database')),
    cfg.StrOpt('engine',
               default='',
               help=_('Database engine')),
]

CONF = cfg.ConfigOpts()
CONF.register_cli_opts(_core_opts)
CONF.register_cli_opts(_db_opts, 'database')
CONF.register_opts(_quota_opts, 'QUOTAS')


def do_alembic_command(config, cmd, *args, **kwargs):
    try:
        getattr(alembic_command, cmd)(config, *args, **kwargs)
    except alembic_util.CommandError as e:
        alembic_util.err(six.text_type(e))


def do_check_migration(config, cmd):
    do_alembic_command(config, 'branches')
    validate_head_file(config)


def add_alembic_subparser(sub, cmd):
    return sub.add_parser(cmd, help=getattr(alembic_command, cmd).__doc__)


def do_upgrade(config, cmd):
    if not CONF.command.revision and not CONF.command.delta:
        raise SystemExit(_('You must provide a revision or relative delta'))

    revision = CONF.command.revision or ''
    if '-' in revision:
        raise SystemExit(_('Negative relative revision (downgrade) not '
                           'supported'))

    delta = CONF.command.delta
    if delta:
        if '+' in revision:
            raise SystemExit(_('Use either --delta or relative revision, '
                               'not both'))
        if delta < 0:
            raise SystemExit(_('Negative delta (downgrade) not supported'))
        revision = '%s+%d' % (revision, delta)

    if not CONF.command.sql:
        run_sanity_checks(config, revision)
    do_alembic_command(config, cmd, revision, sql=CONF.command.sql)


def no_downgrade(config, cmd):
    raise SystemExit(_("Downgrade no longer supported"))


def do_stamp(config, cmd):
    do_alembic_command(config, cmd,
                       CONF.command.revision,
                       sql=CONF.command.sql)


def do_revision(config, cmd):
    do_alembic_command(config, cmd,
                       message=CONF.command.message,
                       autogenerate=CONF.command.autogenerate,
                       sql=CONF.command.sql)
    update_head_file(config)


def validate_head_file(config):
    script = alembic_script.ScriptDirectory.from_config(config)
    if len(script.get_heads()) > 1:
        alembic_util.err(_('Timeline branches unable to generate timeline'))

    head_path = os.path.join(script.versions, HEAD_FILENAME)
    if (os.path.isfile(head_path) and
            open(head_path).read().strip() == script.get_current_head()):
        return
    else:
        alembic_util.err(_('HEAD file does not match migration timeline head'))


def update_head_file(config):
    script = alembic_script.ScriptDirectory.from_config(config)
    if len(script.get_heads()) > 1:
        alembic_util.err(_('Timeline branches unable to generate timeline'))

    head_path = os.path.join(script.versions, HEAD_FILENAME)
    with open(head_path, 'w+') as f:
        f.write(script.get_current_head())


def add_command_parsers(subparsers):
    for name in ['current', 'history', 'branches']:
        parser = add_alembic_subparser(subparsers, name)
        parser.set_defaults(func=do_alembic_command)

    help_text = (getattr(alembic_command, 'branches').__doc__ +
                 ' and validate head file')
    parser = subparsers.add_parser('check_migration', help=help_text)
    parser.set_defaults(func=do_check_migration)

    parser = add_alembic_subparser(subparsers, 'upgrade')
    parser.add_argument('--delta', type=int)
    parser.add_argument('--sql', action='store_true')
    parser.add_argument('revision', nargs='?')
    parser.add_argument('--mysql-engine',
                        default='',
                        help='Change MySQL storage engine of current '
                             'existing tables')
    parser.set_defaults(func=do_upgrade)

    parser = subparsers.add_parser('downgrade', help="(No longer supported)")
    parser.add_argument('None', nargs='?', help="Downgrade not supported")
    parser.set_defaults(func=no_downgrade)

    parser = add_alembic_subparser(subparsers, 'stamp')
    parser.add_argument('--sql', action='store_true')
    parser.add_argument('revision')
    parser.set_defaults(func=do_stamp)

    parser = add_alembic_subparser(subparsers, 'revision')
    parser.add_argument('-m', '--message')
    parser.add_argument('--autogenerate', action='store_true')
    parser.add_argument('--sql', action='store_true')
    parser.set_defaults(func=do_revision)


command_opt = cfg.SubCommandOpt('command',
                                title='Command',
                                help=_('Available commands'),
                                handler=add_command_parsers)

CONF.register_cli_opt(command_opt)


def get_script_location(neutron_config):
    location = '%s.db.migration:alembic_migrations'
    base = "neutron"
    return location % base


def run_sanity_checks(config, revision):
    script_dir = alembic_script.ScriptDirectory.from_config(config)

    def check_sanity(rev, context):
        for script in script_dir.iterate_revisions(revision, rev):
            if hasattr(script.module, 'check_sanity'):
                script.module.check_sanity(context.connection)
        return []

    with environment.EnvironmentContext(config, script_dir,
                                        fn=check_sanity,
                                        starting_rev=None,
                                        destination_rev=revision):
        script_dir.run_env()


def main():
    config = alembic_config.Config(
        os.path.join(os.path.dirname(__file__), '../alembic.ini')
    )
    config.set_main_option('script_location',
                           'quark.db.migration:alembic')
    # attach the Neutron conf to the Alembic conf
    config.neutron_config = CONF
    CONF()
    # TODO(gongysh) enable logging
    CONF.command.func(config, CONF.command.name)



# Copyright 2014 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#

from oslo_log import log as logging

from quark.cache import security_groups_client as sg_client
from quark import environment as env

LOG = logging.getLogger(__name__)


class SecurityGroupDriver(object):
    @env.has_capability(env.Capabilities.SECURITY_GROUPS)
    def update_port(self, **kwargs):
        client = sg_client.SecurityGroupsClient()
        if "security_groups" in kwargs:
            device_id = kwargs.get('device_id')
            mac_address = kwargs.get('mac_address')
            if not device_id or not mac_address:
                LOG.warning('device_id or mac_address not given, ignored.')
                return
            if kwargs["security_groups"]:
                payload = client.serialize_groups(
                    kwargs["security_groups"])
                client.apply_rules(device_id, mac_address, payload)
            else:
                client.delete_vif_rules(device_id, mac_address)

    @env.has_capability(env.Capabilities.SECURITY_GROUPS)
    def delete_port(self, **kwargs):
        client = sg_client.SecurityGroupsClient()
        try:
            device_id = kwargs.get('device_id')
            mac_address = kwargs.get('mac_address')
            if not device_id or not mac_address:
                LOG.warning('device_id or mac_address not given, ignored.')
                return
            client.delete_vif(device_id, mac_address)
        except Exception:
            LOG.exception("Failed to reach the security groups backend")



import contextlib
import datetime

import mock
import netaddr
from neutron.common import exceptions as n_exc_ext
from neutron.common import rpc
from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_utils import timeutils

from quark.db import api as db_api
import quark.ipam
from quark.tests.functional.mysql.base import MySqlBaseFunctionalTest


class QuarkIpamBaseFunctionalTest(MySqlBaseFunctionalTest):
    def setUp(self):
        super(QuarkIpamBaseFunctionalTest, self).setUp()

        patcher = mock.patch("neutron.common.rpc.oslo_messaging")
        patcher.start()
        self.addCleanup(patcher.stop)
        rpc.init(mock.MagicMock())


class QuarkIPAddressReallocate(QuarkIpamBaseFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, network, subnet, address, lock=False):
        self.ipam = quark.ipam.QuarkIpamANY()
        with self.context.session.begin():
            next_ip = subnet.pop("next_auto_assign_ip", 0)
            net_mod = db_api.network_create(self.context, **network)
            subnet["network"] = net_mod
            sub_mod = db_api.subnet_create(self.context, **subnet)

            address["network_id"] = net_mod["id"]
            address["subnet_id"] = sub_mod["id"]
            ip = db_api.ip_address_create(self.context, **address)
            address.pop("address")
            ip = db_api.ip_address_update(self.context, ip, **address)

            # NOTE(asadoughi): update after cidr constructor has been invoked
            db_api.subnet_update(self.context,
                                 sub_mod,
                                 next_auto_assign_ip=next_ip)

        if lock:
            db_api.lock_holder_create(self.context, ip,
                                      name="testlock", type="ip_address")
        yield net_mod

    def test_allocate_finds_ip_reallocates(self):
        network = dict(name="public", tenant_id="fake")
        ipnet = netaddr.IPNetwork("0.0.0.0/24")
        next_ip = ipnet.ipv6().first + 10
        subnet = dict(cidr="0.0.0.0/24", next_auto_assign_ip=next_ip,
                      ip_policy=None, tenant_id="fake", do_not_use=False)

        addr = netaddr.IPAddress("0.0.0.2")

        after_reuse_after = cfg.CONF.QUARK.ipam_reuse_after + 1
        reusable_after = datetime.timedelta(seconds=after_reuse_after)
        deallocated_at = timeutils.utcnow() - reusable_after
        ip_address = dict(address=addr, version=4, _deallocated=True,
                          deallocated_at=deallocated_at)

        with self._stubs(network, subnet, ip_address) as net:
            ipaddress = []
            self.ipam.allocate_ip_address(self.context, ipaddress,
                                          net["id"], 0, 0)
            self.assertIsNotNone(ipaddress[0]['id'])
            expected = netaddr.IPAddress("0.0.0.2").ipv6().value
            self.assertEqual(ipaddress[0]['address'], expected)
            self.assertEqual(ipaddress[0]['version'], 4)
            self.assertEqual(ipaddress[0]['used_by_tenant_id'], "fake")

    def test_allocate_finds_ip_in_do_not_use_subnet_raises(self):
        network = dict(name="public", tenant_id="fake")
        ipnet = netaddr.IPNetwork("0.0.0.0/24")
        next_ip = ipnet.ipv6().first + 3
        subnet = dict(cidr="0.0.0.0/24", next_auto_assign_ip=next_ip,
                      ip_policy=None, tenant_id="fake", do_not_use=True)

        addr = netaddr.IPAddress("0.0.0.2")
        after_reuse_after = cfg.CONF.QUARK.ipam_reuse_after + 1
        reusable_after = datetime.timedelta(seconds=after_reuse_after)
        deallocated_at = timeutils.utcnow() - reusable_after
        ip_address = dict(address=addr, version=4, _deallocated=True,
                          deallocated_at=deallocated_at)

        with self._stubs(network, subnet, ip_address) as net:
            with self.assertRaises(n_exc.IpAddressGenerationFailure):
                self.ipam.allocate_ip_address(self.context, [], net["id"],
                                              0, 0)

    def test_allocate_finds_ip_locked_allocates_next_ip(self):
        network = dict(name="public", tenant_id="fake")
        ipnet = netaddr.IPNetwork("0.0.0.0/24")
        next_ip = ipnet.ipv6().first + 10
        subnet = dict(cidr="0.0.0.0/24", next_auto_assign_ip=next_ip,
                      ip_policy=None, tenant_id="fake", do_not_use=False)

        addr = netaddr.IPAddress("0.0.0.2")

        after_reuse_after = cfg.CONF.QUARK.ipam_reuse_after + 1
        reusable_after = datetime.timedelta(seconds=after_reuse_after)
        deallocated_at = timeutils.utcnow() - reusable_after
        ip_address = dict(address=addr, version=4, _deallocated=True,
                          deallocated_at=deallocated_at)

        with self._stubs(network, subnet, ip_address, lock=True) as net:
            ipaddress = []
            self.ipam.allocate_ip_address(self.context, ipaddress,
                                          net["id"], 0, 0)
            self.assertIsNotNone(ipaddress[0]['id'])
            self.assertEqual(ipaddress[0]['address'], next_ip)
            self.assertEqual(ipaddress[0]['version'], 4)
            self.assertEqual(ipaddress[0]['used_by_tenant_id'], "fake")


class MacAddressReallocate(QuarkIpamBaseFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, do_not_use):
        self.ipam = quark.ipam.QuarkIpamANY()
        mar = db_api.mac_address_range_create(
            self.context,
            cidr="00:00:00:00:00:00/40",
            first_address=0, last_address=255,
            next_auto_assign_mac=6,
            do_not_use=do_not_use)
        mac = db_api.mac_address_create(
            self.context,
            address=1,
            mac_address_range=mar)
        db_api.mac_address_update(
            self.context, mac,
            deallocated=True,
            deallocated_at=datetime.datetime(1970, 1, 1))
        self.context.session.flush()
        yield mar

    def test_reallocate_mac(self):
        with self._stubs(do_not_use=False):
            realloc_mac = self.ipam.allocate_mac_address(self.context, 0, 0, 0)
            self.assertEqual(realloc_mac["address"], 1)

    def test_delete_mac_with_mac_range_do_not_use(self):
        macs = lambda mar: db_api.mac_address_find(
            self.context,
            mac_address_range_id=mar["id"],
            scope=db_api.ALL)
        with self._stubs(do_not_use=True) as mar:
            self.assertEqual(len(macs(mar)), 1)
            with self.assertRaises(n_exc_ext.MacAddressGenerationFailure):
                self.ipam.allocate_mac_address(self.context, 0, 0, 0)
            self.assertEqual(len(macs(mar)), 0)



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#  under the License.

import aiclib

import contextlib

import mock
import neutron.extensions.securitygroup as sg_ext
from oslo_config import cfg

import quark.drivers.nvp_driver
from quark import exceptions as q_exc
from quark.tests import test_base


class TestNVPDriver(test_base.TestBase):
    def setUp(self):
        super(TestNVPDriver, self).setUp()
        cfg.CONF.set_override('environment_capabilities', [], 'QUARK')
        if not hasattr(self, 'driver'):
            self.driver = quark.drivers.nvp_driver.NVPDriver()
        cfg.CONF.clear_override('environment_capabilities', 'QUARK')
        cfg.CONF.set_override('max_rules_per_group', 3, 'NVP')
        cfg.CONF.set_override('max_rules_per_port', 1, 'NVP')
        self.driver.max_ports_per_switch = 0

        self.lswitch_uuid = "12345678-1234-1234-1234-123456781234"
        self.context.tenant_id = "tid"
        self.lport_uuid = "12345678-0000-0000-0000-123456781234"
        self.net_id = "12345678-1234-1234-1234-123412341234"
        self.port_id = "12345678-0000-0000-0000-123412341234"
        self.profile_id = "12345678-0000-0000-0000-000000000000"
        self.d_pkg = "quark.drivers.nvp_driver.NVPDriver"
        self.max_spanning = 3
        self.driver.limits.update({'max_rules_per_group': 3,
                                   'max_rules_per_port': 2})

    def _create_connection(self, switch_count=1,
                           has_switches=False, maxed_ports=False):
        connection = mock.Mock()
        lswitch = self._create_lswitch(has_switches, maxed_ports=maxed_ports)
        lswitchport = self._create_lswitch_port(self.lswitch_uuid,
                                                switch_count)
        connection.lswitch_port = mock.Mock(return_value=lswitchport)
        connection.lswitch = mock.Mock(return_value=lswitch)
        return connection

    def _create_lswitch_port(self, switch_uuid, switch_count):
        port = mock.Mock()
        port.create = mock.Mock(return_value={'uuid': self.lport_uuid})
        port_query = self._create_lport_query(switch_count)
        port.query = mock.Mock(return_value=port_query)
        port.delete = mock.Mock(return_value=None)
        port.attachment_vif = mock.Mock()
        return port

    def _create_lport_query(self, switch_count, profiles=[]):
        query = mock.Mock()
        port_list = {"_relations":
                     {"LogicalSwitchConfig":
                      {"uuid": self.lswitch_uuid,
                       "security_profiles": profiles}}}
        port_query = {"results": [port_list], "result_count": switch_count}
        query.results = mock.Mock(return_value=port_query)
        query.security_profile_uuid().results.return_value = {
            "results": [{"security_profiles": profiles}]}
        return query

    def _create_lswitch(self, switches_available, maxed_ports):
        lswitch = mock.Mock()
        lswitch.query = mock.Mock(
            return_value=self.
            _create_lswitch_query(switches_available, maxed_ports))
        lswitch.create = mock.Mock(return_value={'uuid': self.lswitch_uuid})
        lswitch.delete = mock.Mock(return_value=None)
        return lswitch

    def _create_lswitch_query(self, switches_available, maxed_ports):
        query = mock.Mock()
        port_count = 0
        if maxed_ports:
            port_count = self.max_spanning
        lswitch_list = [{'uuid': 'abcd',
                         '_relations': {
                             'LogicalSwitchStatus': {
                                 'lport_count': port_count}}}]
        if not switches_available:
            lswitch_list = []
        lswitch_query = {"results": lswitch_list}
        query.relations = mock.Mock(return_value=None)
        query.results = mock.Mock(return_value=lswitch_query)
        return query

    def _create_security_profile(self):
        profile = mock.Mock()
        query = mock.Mock()
        group = {'name': 'foo', 'uuid': self.profile_id,
                 'logical_port_ingress_rules': [],
                 'logical_port_egress_rules': []}
        query.results = mock.Mock(return_value={'results': [group],
                                                'result_count': 1})
        profile.query = mock.Mock(return_value=query)
        profile.read = mock.Mock(return_value=group)
        return mock.Mock(return_value=profile)

    def _create_security_rule(self, rule={}):
        return lambda *x, **y: dict(y, ethertype=x[0])


class TestNVPDriverCreateNetwork(TestNVPDriver):
    @contextlib.contextmanager
    def _stubs(self):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
        ) as (conn,):
            connection = self._create_connection()
            conn.return_value = connection
            yield connection

    def test_create_network(self):
        with self._stubs() as (connection):
            self.driver.create_network(self.context, "test")
            self.assertTrue(connection.lswitch().create.called)
            self.assertTrue(connection.lswitch().transport_zone.called)

    def test_create_network_name_longer_than_40_chars_gets_trimmed(self):
        with self._stubs() as (connection):
            long_n = 'A' * 50
            self.driver.create_network(self.context, long_n)
            self.assertTrue(connection.lswitch().create.called)
            self.assertTrue(connection.lswitch().transport_zone.called)
            connection.lswitch().display_name.assert_called_with(long_n[:40])


class TestNVPDriverDefaultTransportZoneBindings(TestNVPDriver):

    def setUp(self):
        super(TestNVPDriverDefaultTransportZoneBindings, self).setUp()
        cfg.CONF.set_override(
            'additional_default_tz_types', ['vxlan'], 'NVP')
        cfg.CONF.set_override(
            'default_tz', 'tz_uuid', 'NVP')
        cfg.CONF.set_override(
            'default_tz_type', 'stt', 'NVP')

    def tearDown(self):
        super(TestNVPDriverDefaultTransportZoneBindings, self).setUp()
        cfg.CONF.clear_override('additional_default_tz_types', 'NVP')
        cfg.CONF.clear_override('default_tz', 'NVP')
        cfg.CONF.clear_override('default_tz_type', 'NVP')

    @contextlib.contextmanager
    def _stubs(self):
        with contextlib.nested(
            mock.patch("quark.drivers.nvp_driver.SA_REGISTRY."
                       "get_strategy"),
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._lswitches_for_network" % self.d_pkg),
        ) as (sa_get_strategy, conn, switch_list):
            connection = self._create_connection()
            conn.return_value = connection

            ret = {"results": [{"uuid": self.lswitch_uuid}]}
            switch_list().results = mock.Mock(return_value=ret)

            sa_strategy = mock.Mock()
            sa_get_strategy.return_value = sa_strategy
            sa_strategy.allocate.return_value = {"id": 123}

            yield sa_get_strategy, sa_strategy, connection

    def test_default_tz_bindings_net_create(self):
        with self._stubs() as (sa_get_strategy, sa_strategy, connection):
            self.driver.create_network(
                self.context, "test", network_id="network_id")

            self.assertTrue(connection.lswitch().create.called)

            # assert vxlan tz manager was called
            sa_strategy.allocate.assert_called_once_with(
                self.context, 'tz_uuid', 'network_id')

            # assert transport_zone was called:
            # once for the default configured tz type (stt)
            # once for the additional default tz type (vxlan)
            self.assertEqual(
                connection.lswitch().transport_zone.call_args_list,
                [mock.call('tz_uuid', 'stt'),
                 mock.call('tz_uuid', 'vxlan', vxlan_id=123)]
            )

    def test_default_tz_bindings_net_delete(self):
        with self._stubs() as (sa_get_strategy, sa_strategy, connection):
            self.driver.delete_network(self.context, "network_id")
            self.assertTrue(connection.lswitch().delete.called)

            sa_strategy.deallocate.assert_called_once_with(
                self.context, 'tz_uuid', 'network_id')


class TestNVPDriverProviderNetwork(TestNVPDriver):
    """Testing all of the network types is unnecessary, but a nice have."""

    @contextlib.contextmanager
    def _stubs(self, tz):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
        ) as (conn,):
            connection = self._create_connection()
            switch = self._create_lswitch(1, False)
            switch.transport_zone = mock.Mock()
            tz_results = mock.Mock()
            tz_results.results = mock.Mock(return_value=tz)
            tz_query = mock.Mock()
            tz_query.query = mock.Mock(return_value=tz_results)
            connection.transportzone = mock.Mock(return_value=tz_query)
            conn.return_value = connection
            yield connection, switch

    def test_config_provider_attrs_flat_net(self):
        tz = dict(result_count=1)
        with self._stubs(tz) as (connection, switch):
            self.driver._config_provider_attrs(
                connection=connection, switch=switch, phys_net="net_uuid",
                net_type="flat", segment_id=None)
            switch.transport_zone.assert_called_with(
                zone_uuid="net_uuid", transport_type="bridge", vlan_id=None)

    def test_config_provider_attrs_vlan_net(self):
        tz = dict(result_count=1)
        with self._stubs(tz) as (connection, switch):
            self.driver._config_provider_attrs(
                connection=connection, switch=switch, phys_net="net_uuid",
                net_type="vlan", segment_id=10)
            switch.transport_zone.assert_called_with(
                zone_uuid="net_uuid", transport_type="bridge", vlan_id=10)

    def test_config_provider_attrs_gre_net(self):
        tz = dict(result_count=1)
        with self._stubs(tz) as (connection, switch):
            self.driver._config_provider_attrs(
                connection=connection, switch=switch, phys_net="net_uuid",
                net_type="gre", segment_id=None)
            switch.transport_zone.assert_called_with(
                zone_uuid="net_uuid", transport_type="gre", vlan_id=None)

    def test_config_provider_attrs_stt_net(self):
        tz = dict(result_count=1)
        with self._stubs(tz) as (connection, switch):
            self.driver._config_provider_attrs(
                connection=connection, switch=switch, phys_net="net_uuid",
                net_type="stt", segment_id=None)
            switch.transport_zone.assert_called_with(
                zone_uuid="net_uuid", transport_type="stt", vlan_id=None)

    def test_config_provider_attrs_local_net(self):
        tz = dict(result_count=1)
        with self._stubs(tz) as (connection, switch):
            self.driver._config_provider_attrs(
                connection=connection, switch=switch, phys_net="net_uuid",
                net_type="local", segment_id=None)
            switch.transport_zone.assert_called_with(
                zone_uuid="net_uuid", transport_type="local", vlan_id=None)

    def test_config_provider_attrs_bridge_net(self):
        """A specialized case for NVP

        This exists because internal driver calls can also call this method,
        and they may pass bridge in as the type as that's how it's known
        to NVP.
        """

        tz = dict(result_count=1)
        with self._stubs(tz) as (connection, switch):
            self.driver._config_provider_attrs(
                connection=connection, switch=switch, phys_net="net_uuid",
                net_type="bridge", segment_id=None)
            switch.transport_zone.assert_called_with(
                zone_uuid="net_uuid", transport_type="bridge", vlan_id=None)

    def test_config_provider_attrs_no_phys_net_or_type(self):
        with self._stubs({}) as (connection, switch):
            self.driver._config_provider_attrs(
                connection=connection, switch=switch, phys_net=None,
                net_type=None, segment_id=None)
            self.assertFalse(switch.transport_zone.called)

    def test_config_provider_attrs_vlan_net_no_segment_id_fails(self):
        with self._stubs({}) as (connection, switch):
            self.assertRaises(
                q_exc.SegmentIdRequired,
                self.driver._config_provider_attrs, connection=connection,
                switch=switch, phys_net="net_uuid", net_type="vlan",
                segment_id=None)

    def test_config_provider_attrs_non_vlan_net_with_segment_id_fails(self):
        with self._stubs({}) as (connection, switch):
            self.assertRaises(
                q_exc.SegmentIdUnsupported,
                self.driver._config_provider_attrs, connection=connection,
                switch=switch, phys_net="net_uuid", net_type="flat",
                segment_id=10)

    def test_config_phys_net_no_phys_type_fails(self):
        with self._stubs({}) as (connection, switch):
            self.assertRaises(
                q_exc.ProvidernetParamError,
                self.driver._config_provider_attrs, connection=connection,
                switch=switch, phys_net="net_uuid", net_type=None,
                segment_id=None)

    def test_config_no_phys_net_with_phys_type_fails(self):
        with self._stubs({}) as (connection, switch):
            self.assertRaises(
                q_exc.ProvidernetParamError,
                self.driver._config_provider_attrs, connection=connection,
                switch=switch, phys_net=None, net_type="flat",
                segment_id=None)

    def test_config_physical_net_doesnt_exist_fails(self):
        tz = dict(result_count=0)
        with self._stubs(tz) as (connection, switch):
            self.assertRaises(
                q_exc.PhysicalNetworkNotFound,
                self.driver._config_provider_attrs, connection=connection,
                switch=switch, phys_net="net_uuid", net_type="flat",
                segment_id=None)

    def test_config_physical_net_bad_net_type_fails(self):
        with self._stubs({}) as (connection, switch):
            self.assertRaises(
                q_exc.InvalidPhysicalNetworkType,
                self.driver._config_provider_attrs, connection=connection,
                switch=switch, phys_net="net_uuid", net_type="lol",
                segment_id=None)


class TestNVPDriverDeleteNetwork(TestNVPDriver):
    @contextlib.contextmanager
    def _stubs(self, network_exists=True):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._lswitches_for_network" % self.d_pkg),
        ) as (conn, switch_list):
            connection = self._create_connection()
            conn.return_value = connection
            if network_exists:
                ret = {"results": [{"uuid": self.lswitch_uuid}]}
            else:
                ret = {"results": []}
            switch_list().results = mock.Mock(return_value=ret)
            yield connection

    def test_delete_network(self):
        with self._stubs() as (connection):
            self.driver.delete_network(self.context, "test")
            self.assertTrue(connection.lswitch().delete.called)

    def test_delete_network_not_exists(self):
        with self._stubs(network_exists=False) as (connection):
            self.driver.delete_network(self.context, "test")
            self.assertFalse(connection.lswitch().delete.called)

    def test_delete_network_not_exists_404_exception(self):
        with self._stubs(network_exists=True) as (connection):
            self.driver.delete_network(self.context, "test")
            self.assertTrue(connection.lswitch().delete.called)


class TestNVPDriverDeleteNetworkWithExceptions(TestNVPDriver):
    @contextlib.contextmanager
    def _stubs(self, network_exists=True, exception=None):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._lswitches_for_network" % self.d_pkg),
            mock.patch("%s._lswitch_delete" % self.d_pkg),
        ) as (conn, switch_list, switch_delete):
            connection = self._create_connection()
            conn.return_value = connection
            if network_exists:
                ret = {"results": [{"uuid": self.lswitch_uuid}]}
            else:
                ret = {"results": []}
            switch_list().results = mock.Mock(return_value=ret)
            if exception:
                switch_delete.side_effect = exception
            yield connection

    def test_delete_network_with_404_aicexception(self):
        e = aiclib.core.AICException(404, 'foo')
        with self._stubs(exception=e) as connection:
            try:
                with self.assertRaises(type(e)):
                    self.driver.delete_network(self.context, "test")
                self.fail("AssertionError should have been raised.")
            except AssertionError as ae:
                self.assertEqual(ae.args[0], "AICException not raised")
                self.assertFalse(connection.lswitch().delete.called)

    def test_delete_network_with_500_aicexception(self):
        e = aiclib.core.AICException(500, 'foo')
        with self._stubs(exception=e) as connection:
            try:
                with self.assertRaises(type(e)):
                    self.driver.delete_network(self.context, "test")
                self.fail("AssertionError should have been raised.")
            except AssertionError as ae:
                self.assertEqual(ae.args[0], "AICException not raised")
                self.assertFalse(connection.lswitch().delete.called)

    def test_delete_network_with_normal_exception(self):
        e = StandardError('foo')
        with self._stubs(exception=e) as connection:
            try:
                with self.assertRaises(type(e)):
                    self.driver.delete_network(self.context, "test")
                self.fail("AssertionError should have been raised.")
            except AssertionError as ae:
                self.assertEqual(ae.args[0], "StandardError not raised")
                self.assertFalse(connection.lswitch().delete.called)


class TestNVPDriverCreatePort(TestNVPDriver):
    '''In all cases an lswitch should be queried.'''
    @contextlib.contextmanager
    def _stubs(self, has_lswitch=True, maxed_ports=False, net_details=None):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._next_connection" % self.d_pkg),
            mock.patch("%s._lswitches_for_network" % self.d_pkg),
            mock.patch("%s._get_network_details" % self.d_pkg),
        ) as (conn, next_conn, get_switches, get_net_dets):
            connection = self._create_connection(has_switches=has_lswitch,
                                                 maxed_ports=maxed_ports)
            conn.return_value = connection
            get_switches.return_value = connection.lswitch().query()
            get_net_dets.return_value = net_details
            yield connection

    def test_select_ipam_strategy(self):
        strategy = self.driver.select_ipam_strategy(1, "ANY")
        self.assertEqual(strategy, "ANY")

    def test_create_port_switch_exists(self):
        with self._stubs(net_details=dict(foo=3)) as (connection):
            port = self.driver.create_port(self.context, self.net_id,
                                           self.port_id)
            self.assertTrue("uuid" in port)
            self.assertTrue(connection.lswitch_port().attachment_vif.called)
            self.assertFalse(connection.lswitch().create.called)
            self.assertTrue(connection.lswitch_port().create.called)
            self.assertTrue(connection.lswitch().query.called)
            status_args, kwargs = (
                connection.lswitch_port().admin_status_enabled.call_args)
            self.assertTrue(True in status_args)

    def test_create_port_switch_exists_tags(self):
        with self._stubs(net_details=dict(foo=3)) as (connection):
            device_id = "foo"
            port = self.driver.create_port(self.context, self.net_id,
                                           self.port_id, device_id=device_id)
            self.assertTrue("uuid" in port)
            self.assertTrue(connection.lswitch_port().attachment_vif.called)
            self.assertFalse(connection.lswitch().create.called)
            self.assertTrue(connection.lswitch_port().create.called)
            self.assertTrue(connection.lswitch().query.called)
            status_args, kwargs = (
                connection.lswitch_port().admin_status_enabled.call_args)
            self.assertTrue(True in status_args)
            connection.lswitch_port().assert_has_calls([mock.call.tags([
                dict(scope="neutron_net_id", tag=self.net_id),
                dict(scope="neutron_port_id", tag=self.port_id),
                dict(scope="os_tid", tag=self.context.tenant_id),
                dict(scope="vm_id", tag=device_id)
            ])], any_order=True)

    def test_create_port_switch_not_exists(self):
        with self._stubs(has_lswitch=False,
                         net_details=dict(foo=3)) as (connection):
            port = self.driver.create_port(self.context, self.net_id,
                                           self.port_id)
            self.assertTrue("uuid" in port)
            self.assertTrue(connection.lswitch().create.called)
            self.assertTrue(connection.lswitch_port().create.called)
            self.assertTrue(connection.lswitch().query.called)
            status_args, kwargs = (
                connection.lswitch_port().admin_status_enabled.call_args)
            self.assertTrue(True in status_args)

    def test_create_port_no_existing_switches_fails(self):
        with self._stubs(has_lswitch=False):
            self.assertRaises(q_exc.BadNVPState, self.driver.create_port,
                              self.context, self.net_id, self.port_id, False)

    def test_create_disabled_port_switch_not_exists(self):
        with self._stubs(has_lswitch=False,
                         net_details=dict(foo=3)) as (connection):
            port = self.driver.create_port(self.context, self.net_id,
                                           self.port_id, False)
            self.assertTrue("uuid" in port)
            self.assertTrue(connection.lswitch().create.called)
            self.assertTrue(connection.lswitch_port().create.called)
            self.assertTrue(connection.lswitch().query.called)
            status_args, kwargs = (
                connection.lswitch_port().admin_status_enabled.call_args)
            self.assertTrue(False in status_args)

    def test_create_port_switch_exists_spanning(self):
        with self._stubs(maxed_ports=True,
                         net_details=dict(foo=3)) as (connection):
            self.driver.limits['max_ports_per_switch'] = self.max_spanning
            port = self.driver.create_port(self.context, self.net_id,
                                           self.port_id)
            self.assertTrue("uuid" in port)
            self.assertTrue(connection.lswitch().create.called)
            self.assertTrue(connection.lswitch_port().create.called)
            self.assertTrue(connection.lswitch().query.called)
            status_args, kwargs = (
                connection.lswitch_port().admin_status_enabled.call_args)
            self.assertTrue(True in status_args)

    def test_create_port_switch_not_exists_spanning(self):
        with self._stubs(has_lswitch=False, maxed_ports=True,
                         net_details=dict(foo=3)) as (connection):
            self.driver.max_ports_per_switch = self.max_spanning
            port = self.driver.create_port(self.context, self.net_id,
                                           self.port_id)
            self.assertTrue("uuid" in port)
            self.assertTrue(connection.lswitch().create.called)
            self.assertTrue(connection.lswitch_port().create.called)
            self.assertTrue(connection.lswitch().query.called)
            status_args, kwargs = (
                connection.lswitch_port().admin_status_enabled.call_args)
            self.assertTrue(True in status_args)

    def test_create_disabled_port_switch_not_exists_spanning(self):
        with self._stubs(has_lswitch=False, maxed_ports=True,
                         net_details=dict(foo=3)) as (connection):
            self.driver.max_ports_per_switch = self.max_spanning
            port = self.driver.create_port(self.context, self.net_id,
                                           self.port_id, False)
            self.assertTrue("uuid" in port)
            self.assertTrue(connection.lswitch().create.called)
            self.assertTrue(connection.lswitch_port().create.called)
            self.assertTrue(connection.lswitch().query.called)
            status_args, kwargs = (
                connection.lswitch_port().admin_status_enabled.call_args)
            self.assertTrue(False in status_args)

    def test_create_port_with_security_groups(self):
        cfg.CONF.set_override('environment_capabilities', [], 'QUARK')
        with self._stubs() as connection:
            connection.securityprofile = self._create_security_profile()
            self.driver.create_port(self.context, self.net_id,
                                    self.port_id,
                                    security_groups=[1])
            connection.lswitch_port().assert_has_calls([
                mock.call.security_profiles([self.profile_id]),
            ], any_order=True)
        cfg.CONF.clear_override('environment_capabilities', 'QUARK')

    def test_create_port_with_security_groups_max_rules(self):
        cfg.CONF.set_override('environment_capabilities', [], 'QUARK')
        with self._stubs() as connection:
            connection.securityprofile = self._create_security_profile()
            connection.securityprofile().read().update(
                {'logical_port_ingress_rules': [{'ethertype': 'IPv4'},
                                                {'ethertype': 'IPv6'}],
                 'logical_port_egress_rules': [{'ethertype': 'IPv4'},
                                               {'ethertype': 'IPv6'}]})
            with self.assertRaises(sg_ext.nexception.InvalidInput):
                self.driver.create_port(
                    self.context, self.net_id, self.port_id,
                    security_groups=[1])
        cfg.CONF.clear_override('environment_capabilities', 'QUARK')


class TestNVPDriverUpdatePort(TestNVPDriver):
    @contextlib.contextmanager
    def _stubs(self):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._next_connection" % self.d_pkg),
        ) as (conn, next_conn):
            connection = self._create_connection()
            connection.securityprofile = self._create_security_profile()
            conn.return_value = connection
            yield connection

    def test_update_port(self):
        cfg.CONF.set_override('environment_capabilities', [], 'QUARK')
        with self._stubs() as connection:
            self.driver.update_port(
                self.context, self.port_id,
                security_groups=[1])
            connection.lswitch_port().assert_has_calls([
                mock.call.security_profiles([self.profile_id]),
            ], any_order=True)
        cfg.CONF.clear_override('environment_capabilities', 'QUARK')

    def test_update_port_max_rules(self):
        cfg.CONF.set_override('environment_capabilities', [], 'QUARK')
        with self._stubs() as connection:
            connection.securityprofile().read().update(
                {'logical_port_ingress_rules': [{'ethertype': 'IPv4'},
                                                {'ethertype': 'IPv6'}],
                 'logical_port_egress_rules': [{'ethertype': 'IPv4'},
                                               {'ethertype': 'IPv6'}]})
            with self.assertRaises(sg_ext.nexception.InvalidInput):
                self.driver.update_port(
                    self.context, self.port_id,
                    security_groups=[1])
        cfg.CONF.clear_override('environment_capabilities', 'QUARK')


class TestNVPDriverLswitchesForNetwork(TestNVPDriver):
    @contextlib.contextmanager
    def _stubs(self, single_switch=True):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
        ) as (conn,):
            connection = self._create_connection(
                has_switches=True, switch_count=1)
            conn.return_value = connection
            yield connection

    def test_get_lswitches(self):
        """Test exists for coverage. No decisions are made."""
        with self._stubs() as connection:
            query_mock = mock.Mock()
            query_mock.tags = mock.Mock()
            query_mock.tagscopes = mock.Mock()
            connection.query = mock.Mock(return_value=query_mock)
            self.driver._lswitches_for_network(self.context, "net_uuid")

    def test_get_lswitch_ids_for_network(self):
        with self._stubs() as connection:
            query_mock = mock.Mock()
            query_mock.tags = mock.Mock()
            query_mock.tagscopes = mock.Mock()
            connection.query = mock.Mock(return_value=query_mock)
            lswitch_ids = self.driver.get_lswitch_ids_for_network(
                self.context, "net_uuid")
            self.assertEqual(lswitch_ids, ['abcd'])


class TestSwitchCopying(TestNVPDriver):
    def test_no_existing_switches(self):
        switches = dict(results=[])
        args = self.driver._get_network_details(None, 1, switches)
        self.assertTrue(args == {})

    def test_has_switches_no_transport_zones(self):
        switch = dict(display_name="public", transport_zones=[])
        switches = dict(results=[switch])
        args = self.driver._get_network_details(None, 1, switches)
        self.assertEqual(args["network_name"], "public")
        self.assertEqual(args["phys_net"], None)

    def test_has_switches_and_transport_zones(self):
        transport_zones = [dict(zone_uuid="zone_uuid",
                                transport_type="bridge")]
        switch = dict(display_name="public", transport_zones=transport_zones)
        switches = dict(results=[switch])
        args = self.driver._get_network_details(None, 1, switches)
        self.assertEqual(args["network_name"], "public")
        self.assertEqual(args["phys_net"], "zone_uuid")
        self.assertEqual(args["phys_type"], "bridge")

    def test_has_switches_tz_and_vlan(self):
        binding = dict(vlan_translation=[dict(transport=10)])
        transport_zones = [dict(zone_uuid="zone_uuid",
                                transport_type="bridge",
                                binding_config=binding)]
        switch = dict(display_name="public", transport_zones=transport_zones)
        switches = dict(results=[switch])
        args = self.driver._get_network_details(None, 1, switches)
        self.assertEqual(args["network_name"], "public")
        self.assertEqual(args["phys_net"], "zone_uuid")
        self.assertEqual(args["phys_type"], "bridge")


class TestNVPDriverDeletePort(TestNVPDriver):
    @contextlib.contextmanager
    def _stubs(self, switch_count=1):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._next_connection" % self.d_pkg),
        ) as (conn, next_conn):
            connection = self._create_connection(switch_count=switch_count)
            conn.return_value = connection
            yield connection

    def test_delete_port(self):
        with self._stubs() as (connection):
            self.driver.delete_port(self.context, self.port_id)
            self.assertTrue(connection.lswitch_port().delete.called)

    def test_delete_port_switch_given(self):
        with self._stubs() as (connection):
            self.driver.delete_port(self.context, self.port_id,
                                    lswitch_uuid=self.lswitch_uuid)
            self.assertFalse(connection.lswitch_port().query.called)
            self.assertTrue(connection.lswitch_port().delete.called)

    def test_delete_port_many_switches(self):
        with self._stubs(switch_count=2):
            try:
                with self.assertRaises(Exception):  # noqa
                    self.driver.delete_port(self.context, self.port_id)
            except AssertionError as ae:
                self.assertEqual(ae.args[0], "Exception not raised")

    def test_delete_port_no_switch_bad_data(self):
        with self._stubs(switch_count=0):
            try:
                with self.assertRaises(Exception):  # noqa
                    self.driver.delete_port(self.context, self.port_id)
            except AssertionError as ae:
                self.assertEqual(ae.args[0], "Exception not raised")


class TestNVPDriverDeletePortWithExceptions(TestNVPDriver):
    @contextlib.contextmanager
    def _stubs(self, switch_exception=None, delete_exception=None):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._lswitch_from_port" % self.d_pkg),
        ) as (conn, switch):
            connection = self._create_connection()
            conn.return_value = connection
            if switch_exception:
                switch.side_effect = switch_exception
            else:
                switch = mock.Mock(return_value=1)
            if delete_exception:
                connection.lswitch_port.delete.side_effect = delete_exception
            yield connection

    def test_delete_port_with_switch_query_general_exception(self):
        e = Exception('foo')
        with self._stubs(switch_exception=e) as (connection):
            try:
                with self.assertRaises(type(e)):
                    self.driver.delete_port(self.context, 'test')
                self.fail("AssertionError should have been raised.")
            except AssertionError as ae:
                self.assertEqual(ae.args[0], "Exception not raised")
                self.assertFalse(connection.lswitch_port().delete.called)

    def test_delete_port_with_switch_query_404_aic_exception(self):
        e = aiclib.core.AICException(404, 'foo')
        with self._stubs(switch_exception=e) as (connection):
            try:
                with self.assertRaises(type(e)):
                    self.driver.delete_port(self.context, 'test')
                self.fail("AssertionError should have been raised.")
            except AssertionError as ae:
                self.assertEqual(ae.args[0], "AICException not raised")
                self.assertFalse(connection.lswitch_port().delete.called)

    def test_delete_port_with_switch_query_500_aic_exception(self):
        e = aiclib.core.AICException(500, 'foo')
        with self._stubs(switch_exception=e) as (connection):
            try:
                with self.assertRaises(type(e)):
                    self.driver.delete_port(self.context, 'test')
                self.fail("AssertionError should have been raised.")
            except AssertionError as ae:
                self.assertEqual(ae.args[0], "AICException not raised")
                self.assertFalse(connection.lswitch_port().delete.called)

    def test_delete_port_with_delete_general_exception(self):
        e = Exception('foo')
        with self._stubs(delete_exception=e) as (connection):
            try:
                with self.assertRaises(type(e)):
                    self.driver.delete_port(self.context, 'test')
                self.fail("AssertionError should have been raised.")
            except AssertionError as ae:
                self.assertEqual(ae.args[0], "Exception not raised")
                self.assertTrue(connection.lswitch_port().delete.called)

    def test_delete_port_with_delete_404_aic_exception(self):
        e = aiclib.core.AICException(404, 'foo')
        with self._stubs(delete_exception=e) as (connection):
            try:
                with self.assertRaises(type(e)):
                    self.driver.delete_port(self.context, 'test')
                self.fail("AssertionError should have been raised.")
            except AssertionError as ae:
                self.assertEqual(ae.args[0], "AICException not raised")
                self.assertTrue(connection.lswitch_port().delete.called)

    def test_delete_port_with_delete_500_aic_exception(self):
        e = aiclib.core.AICException(500, 'foo')
        with self._stubs(delete_exception=e) as (connection):
            try:
                with self.assertRaises(type(e)):
                    self.driver.delete_port(self.context, 'test')
                self.fail("AssertionError should have been raised.")
            except AssertionError as ae:
                self.assertEqual(ae.args[0], "AICException not raised")
                self.assertTrue(connection.lswitch_port().delete.called)


class TestNVPDriverCreateSecurityGroup(TestNVPDriver):
    @contextlib.contextmanager
    def _stubs(self):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._next_connection" % self.d_pkg),
        ) as (conn, next_conn):
            connection = self._create_connection()
            connection.securityprofile = self._create_security_profile()
            conn.return_value = connection
            yield connection

    def test_security_group_create(self):
        group = {'group_id': 1}
        with self._stubs() as connection:
            self.driver.create_security_group(
                self.context, 'foo', **group)
            connection.securityprofile().assert_has_calls([
                mock.call.display_name('foo'),
                mock.call.create(),
            ], any_order=True)

    def test_security_group_create_with_rules(self):
        ingress_rules = [{'ethertype': 'IPv4'}, {'ethertype': 'IPv4',
                                                 'protocol': 6}]
        egress_rules = [{'ethertype': 'IPv6', 'protocol': 17}]
        group = {'group_id': 1, 'port_ingress_rules': ingress_rules,
                 'port_egress_rules': egress_rules}
        with self._stubs() as connection:
            self.driver.create_security_group(
                self.context, 'foo', **group)
            connection.securityprofile().assert_has_calls([
                mock.call.display_name('foo'),
                mock.call.port_egress_rules(egress_rules),
                mock.call.port_ingress_rules(ingress_rules),
                mock.call.tags([{'scope': 'neutron_group_id', 'tag': 1},
                                {'scope': 'os_tid',
                                 'tag': self.context.tenant_id}]),
            ], any_order=True)

    def test_security_group_create_rules_at_max(self):
        ingress_rules = [{'ethertype': 'IPv4', 'protocol': 6},
                         {'ethertype': 'IPv6',
                          'remote_ip_prefix': '192.168.0.1'}]
        egress_rules = [{'ethertype': 'IPv4', 'protocol': 17,
                         'port_range_min': 0, 'port_range_max': 100},
                        {'ethertype': 'IPv4', 'remote_group_id': 2}]
        with self._stubs():
            with self.assertRaises(sg_ext.nexception.InvalidInput):
                self.driver.create_security_group(
                    self.context, 'foo',
                    port_ingress_rules=ingress_rules,
                    port_egress_rules=egress_rules)


class TestNVPDriverDeleteSecurityGroup(TestNVPDriver):
    @contextlib.contextmanager
    def _stubs(self):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._next_connection" % self.d_pkg),
        ) as (conn, next_conn):
            connection = self._create_connection()
            connection.securityprofile = self._create_security_profile()
            conn.return_value = connection
            yield connection

    def test_security_group_delete(self):
        with self._stubs() as connection:
            self.driver.delete_security_group(self.context, 1)
            connection.securityprofile().query().assert_has_calls([
                mock.call.tagscopes(['os_tid', 'neutron_group_id']),
                mock.call.tags([self.context.tenant_id, 1]),
            ], any_order=True)
            connection.securityprofile.assert_any_call(self.profile_id)
            self.assertTrue(connection.securityprofile().delete)

    def test_security_group_delete_not_found(self):
        with self._stubs() as connection:
            connection.securityprofile().query().results.return_value = {
                'result_count': 0, 'results': []}
            with self.assertRaises(sg_ext.SecurityGroupNotFound):
                self.driver.delete_security_group(self.context, 1)


class TestNVPDriverUpdateSecurityGroup(TestNVPDriver):
    @contextlib.contextmanager
    def _stubs(self):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._next_connection" % self.d_pkg),
        ) as (conn, next_conn):
            connection = self._create_connection()
            connection.securityprofile = self._create_security_profile()
            conn.return_value = connection
            yield connection

    def test_security_group_update(self):
        with self._stubs() as connection:
            self.driver.update_security_group(self.context, 1, name='bar')
            connection.securityprofile.assert_any_call(self.profile_id)
            connection.securityprofile().assert_has_calls([
                mock.call.display_name('bar'),
                mock.call.update()],
                any_order=True)

    def test_security_group_update_not_found(self):
        with self._stubs() as connection:
            connection.securityprofile().query().results.return_value = {
                'result_count': 0, 'results': []}
            with self.assertRaises(sg_ext.SecurityGroupNotFound):
                self.driver.update_security_group(self.context, 1)

    def test_security_group_update_with_rules(self):
        ingress_rules = [{'ethertype': 'IPv4', 'protocol': 6},
                         {'ethertype': 'IPv6',
                          'remote_ip_prefix': '192.168.0.1'}]
        egress_rules = [{'ethertype': 'IPv4', 'protocol': 17,
                         'port_range_min': 0, 'port_range_max': 100}]
        with self._stubs() as connection:
            self.driver.update_security_group(
                self.context, 1,
                port_ingress_rules=ingress_rules,
                port_egress_rules=egress_rules)
            connection.securityprofile.assert_any_call(self.profile_id)
            connection.securityprofile().assert_has_calls([
                mock.call.port_ingress_rules(ingress_rules),
                mock.call.port_egress_rules(egress_rules),
                mock.call.update(),
            ], any_order=True)

    def test_security_group_update_rules_at_max(self):
        ingress_rules = [{'ethertype': 'IPv4', 'protocol': 6},
                         {'ethertype': 'IPv6',
                          'remote_ip_prefix': '192.168.0.1'}]
        egress_rules = [{'ethertype': 'IPv4', 'protocol': 17,
                         'port_range_min': 0, 'port_range_max': 100},
                        {'ethertype': 'IPv4', 'remote_group_id': 2}]
        with self._stubs():
            with self.assertRaises(sg_ext.nexception.InvalidInput):
                self.driver.update_security_group(
                    self.context, 1,
                    port_ingress_rules=ingress_rules,
                    port_egress_rules=egress_rules)


class TestNVPDriverCreateSecurityGroupRule(TestNVPDriver):
    @contextlib.contextmanager
    def _stubs(self):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._next_connection" % self.d_pkg),
        ) as (conn, next_conn):
            connection = self._create_connection()
            connection.securityprofile = self._create_security_profile()
            connection.securityrule = self._create_security_rule()
            connection.lswitch_port().query.return_value = (
                self._create_lport_query(1, [self.profile_id]))
            conn.return_value = connection
            yield connection

    def test_security_rule_create(self):
        with self._stubs() as connection:
            self.driver.create_security_group_rule(
                self.context, 1,
                {'ethertype': 'IPv4', 'direction': 'ingress'})
            connection.securityprofile.assert_any_call(self.profile_id)
            connection.securityprofile().assert_has_calls([
                mock.call.port_ingress_rules([{'ethertype': 'IPv4'}]),
                mock.call.update(),
            ], any_order=True)

    def test_security_rule_create_with_ip_prefix_and_profile(self):
        with self._stubs() as connection:
            self.driver.create_security_group_rule(
                self.context, 1,
                {'ethertype': 'IPv4', 'direction': 'ingress',
                 'remote_ip_prefix': "pre", "remote_group_id": "group",
                 "protocol": "udp"})
            connection.securityprofile.assert_any_call(self.profile_id)
            connection.securityprofile().assert_has_calls([
                mock.call.port_ingress_rules([{'ethertype': 'IPv4',
                                               "ip_prefix": "pre",
                                               "profile_uuid": "group",
                                               "protocol": "udp"}]),
                mock.call.update(),
            ], any_order=True)

    def test_security_rule_create_invalid_direction(self):
        with self._stubs():
            with self.assertRaises(AttributeError):
                self.driver.create_security_group_rule(
                    self.context, 1,
                    {'ethertype': 'IPv4', 'direction': 'instantregret'})

    def test_security_rule_create_duplicate(self):
        with self._stubs() as connection:
            connection.securityprofile().read().update({
                'logical_port_ingress_rules': [{'ethertype': 'IPv4'}],
                'logical_port_egress_rules': []})
            with self.assertRaises(sg_ext.SecurityGroupRuleExists):
                self.driver.create_security_group_rule(
                    self.context, 1,
                    {'ethertype': 'IPv4', 'direction': 'ingress'})

    def test_security_rule_create_not_found(self):
        with self._stubs() as connection:
            connection.securityprofile().query().results.return_value = {
                'result_count': 0, 'results': []}
            with self.assertRaises(sg_ext.SecurityGroupNotFound):
                self.driver.create_security_group_rule(
                    self.context, 1,
                    {'ethertype': 'IPv4', 'direction': 'egress'})

    def test_security_rule_create_over_port(self):
        with self._stubs() as connection:
            connection.securityprofile().read().update(
                {'logical_port_ingress_rules': [1, 2]})
            with self.assertRaises(sg_ext.nexception.InvalidInput):
                self.driver.create_security_group_rule(
                    self.context, 1,
                    {'ethertype': 'IPv4', 'direction': 'egress'})
            self.assertTrue(connection.lswitch_port().query.called)


class TestNVPDriverDeleteSecurityGroupRule(TestNVPDriver):
    @contextlib.contextmanager
    def _stubs(self, rules=[]):
        rulelist = {'logical_port_ingress_rules': [],
                    'logical_port_egress_rules': []}
        for rule in rules:
            rulelist['logical_port_%s_rules' % rule.pop('direction')].append(
                rule)
        with contextlib.nested(
                mock.patch("%s._connection" % self.d_pkg),
        ) as (conn,):
            connection = self._create_connection()
            connection.securityprofile = self._create_security_profile()
            connection.securityrule = self._create_security_rule()
            connection.securityprofile().read().update(rulelist)
            conn.return_value = connection
            yield connection

    def test_delete_security_group(self):
        with self._stubs(
            rules=[{'ethertype': 'IPv4', 'direction': 'ingress'},
                   {'ethertype': 'IPv6', 'direction': 'egress'}]
        ) as connection:
            self.driver.delete_security_group_rule(
                self.context, 1, {'ethertype': 'IPv6', 'direction': 'egress'})
            connection.securityprofile.assert_any_call(self.profile_id)
            connection.securityprofile().assert_has_calls([
                mock.call.port_egress_rules([]),
                mock.call.update(),
            ], any_order=True)

    def test_delete_security_group_does_not_exist(self):
        with self._stubs(rules=[{'ethertype': 'IPv4',
                                 'direction': 'ingress'}]):
            with self.assertRaises(sg_ext.SecurityGroupRuleNotFound):
                self.driver.delete_security_group_rule(
                    self.context, 1,
                    {'ethertype': 'IPv6', 'direction': 'egress'})


class TestNVPDriverLoadConfig(TestNVPDriver):
    def test_load_config(self):
        controllers = "192.168.221.139:443:admin:admin:30:10:2:2"
        cfg.CONF.set_override("controller_connection", [controllers], "NVP")
        self.driver.load_config()
        conn = self.driver.nvp_connections[0]
        self.assertEqual(conn["username"], "admin")
        self.assertEqual(conn["retries"], 2)
        self.assertEqual(conn["redirects"], '2')
        self.assertEqual(conn["http_timeout"], 10)
        self.assertEqual(conn["req_timeout"], "30")
        self.assertEqual(conn["default_tz"], None)
        self.assertEqual(conn["password"], "admin")
        self.assertEqual(conn["ip_address"], "192.168.221.139")
        self.assertEqual(conn["port"], "443")
        cfg.CONF.clear_override("controller_connection", "NVP")

    def test_load_config_no_connections(self):
        self.driver.load_config()
        self.assertEqual(len(self.driver.nvp_connections), 0)


class TestNVPDriverLoadConfigRandomController(TestNVPDriver):
    @mock.patch("random.randint")
    def test_load_config(self, randint):
        controllers = "192.168.221.139:443:admin:admin:30:10:2:2"
        cfg.CONF.set_override("controller_connection", [controllers], "NVP")
        cfg.CONF.set_override("random_initial_controller", True,
                              "NVP")
        randint.return_value = 0
        self.driver.load_config()
        self.assertTrue(randint.called)
        cfg.CONF.clear_override("controller_connection", "NVP")
        cfg.CONF.clear_override("random_initial_controller", "NVP")


class TestNVPGetConnection(TestNVPDriver):
    @contextlib.contextmanager
    def _stubs(self, has_conn):
        controllers = "192.168.221.139:443:admin:admin:30:10:2:2"
        cfg.CONF.set_override("controller_connection", [controllers], "NVP")
        if has_conn:
            self.driver.nvp_connections.append(dict(connection="foo",
                                                    usages=0))
        else:
            self.driver.nvp_connections.append(dict(port="443",
                                                    ip_address="192.168.0.1",
                                                    username="admin",
                                                    password="admin",
                                                    http_timeout=10,
                                                    retries=1,
                                                    backoff=0,
                                                    usages=0))
        with contextlib.nested(
            mock.patch("aiclib.nvp.Connection"),
            mock.patch("%s._next_connection" % self.d_pkg)
        ) as (aiclib_conn, next_conn):
            yield aiclib_conn, next_conn
        cfg.CONF.clear_override("controller_connection", "NVP")

    def test_get_connection(self):
        with self._stubs(has_conn=False) as (aiclib_conn, next_conn):
            with self.driver.get_connection():
                pass
            self.assertTrue(aiclib_conn.called)
            self.assertFalse(next_conn.called)

    def test_get_connection_connection_defined(self):
        with self._stubs(has_conn=True) as (aiclib_conn, next_conn):
            with self.driver.get_connection():
                pass
            self.assertFalse(aiclib_conn.called)
            self.assertFalse(next_conn.called)

    def test_get_connection_iterates(self):
        with self._stubs(has_conn=True) as (aiclib_conn, next_conn):
            try:
                with self.driver.get_connection():
                    raise Exception("Failure")
            except Exception:
                pass
            self.assertFalse(aiclib_conn.called)
            self.assertTrue(next_conn.called)

    def test_get_connection_with_threshold(self):
        cfg.CONF.set_override("connection_switching_threshold", 1, "NVP")
        with self._stubs(has_conn=True) as (aiclib_conn, next_conn):
            with self.driver.get_connection():
                pass

            self.assertFalse(aiclib_conn.called)
            self.assertTrue(next_conn.called)
        cfg.CONF.clear_override("connection_switching_threshold", "NVP")

    def test_get_connection_with_threshold_next_conn_not_called(self):
        cfg.CONF.set_override("connection_switching_threshold", 2, "NVP")
        with self._stubs(has_conn=True) as (aiclib_conn, next_conn):
            with self.driver.get_connection():
                pass

            self.assertFalse(aiclib_conn.called)
            self.assertFalse(next_conn.called)
        cfg.CONF.clear_override("connection_switching_threshold", "NVP")


class TestNVPGetConnectionNoneDefined(TestNVPDriver):
    def test_get_connection(self):
        with self.assertRaises(q_exc.NoBackendConnectionsDefined):
            with self.driver.get_connection():
                pass


class TestNVPNextConnection(TestNVPDriver):
    @contextlib.contextmanager
    def _stubs(self, rand=False):
        controllers = "192.168.221.139:443:admin:admin:30:10:2:2"
        cfg.CONF.set_override("controller_connection", [controllers], "NVP")
        if rand:
            cfg.CONF.set_override("connection_switching_random", True, "NVP")
        conn1 = dict(port="443", ip_address="192.168.0.1", username="admin",
                     password="admin", http_timeout=10, retries=1, backoff=0,
                     usages=0)
        conn2 = conn1.copy()
        conn2["ip_address"] = "192.168.0.2"

        self.driver.nvp_connections.extend([conn1, conn2])
        with contextlib.nested(
            mock.patch("random.randint")
        ) as (randint,):
            randint.return_value = 1
            yield randint
        cfg.CONF.clear_override("controller_connection", "NVP")
        if rand:
            cfg.CONF.clear_override("connection_switching_random", "NVP")

    def test_get_connection(self):
        with self._stubs() as randint:
            self.driver._next_connection()
            self.assertEqual(1, self.driver.conn_index)
            self.assertFalse(randint.called)

    def test_get_connection_random(self):
        with self._stubs(rand=True) as randint:
            self.driver._next_connection()
            self.assertEqual(1, self.driver.conn_index)
            self.assertTrue(randint.called)



# Copyright (c) 2012 Rackspace Hosting Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sqlalchemy.dialects import sqlite
from sqlalchemy import types


class INET(types.TypeDecorator):
    impl = types.CHAR

    def load_dialect_impl(self, dialect):
        # IPv6 is 128 bits => 2^128 == 3.4e38 => 39 digits
        return dialect.type_descriptor(types.CHAR(39))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        return str(value)

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        return long(value)

    def coerce_compared_value(self, op, value):
        # NOTE(mdietz): If left unimplemented, the column is coerced into a
        # string every time, causing the next_auto_assign_increment to be a
        # string concatenation rather than an addition. 'value' in the
        # signature is the "other" value being compared for the purposes of
        # casting.
        if isinstance(value, int):
            return types.Integer()
        return self


class MACAddress(types.TypeDecorator):
    impl = types.BigInteger

    def load_dialect_impl(self, dialect):
        if dialect.name == 'sqlite':
            return dialect.type_descriptor(sqlite.CHAR)
        return dialect.type_descriptor(self.impl)



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log as logging

from quark import billing
from quark.db import api as db_api
from quark.db import ip_types
from quark.drivers import floating_ip_registry as registry
from quark import exceptions as q_exc
from quark import ipam
from quark import plugin_views as v


CONF = cfg.CONF
LOG = logging.getLogger(__name__)

quark_router_opts = [
    cfg.StrOpt('floating_ip_segment_name', default='floating_ip',
               help=_('Segment name for floating IP subnets')),
    cfg.StrOpt('floating_ip_ipam_strategy', default='ANY',
               help=_('Override the network IPAM stategy for floating '
                      "allocation. Use 'NETWORK' to fall back to the "
                      "network's strategy")),
]

CONF.register_opts(quark_router_opts, 'QUARK')


def _get_network(context, network_id):
    network = db_api.network_find(context, id=network_id, scope=db_api.ONE)
    if not network:
        raise n_exc.NetworkNotFound(net_id=network_id)
    return network


def _get_port(context, port_id):
    port = db_api.port_find(context, id=port_id, scope=db_api.ONE)
    if not port:
        raise n_exc.PortNotFound(port_id=port_id)

    if not port.ip_addresses or len(port.ip_addresses) == 0:
        raise q_exc.NoAvailableFixedIpsForPort(port_id=port_id)
    return port


def _get_fixed_ip(context, given_fixed_ip, port):
    if not given_fixed_ip:
        fixed_ip = _get_next_available_fixed_ip(port)
        if not fixed_ip:
            raise q_exc.NoAvailableFixedIpsForPort(
                port_id=port.id)
    else:
        fixed_ip = next((ip for ip in port.ip_addresses
                        if (ip['address_readable'] == given_fixed_ip and
                            ip.get('address_type') == ip_types.FIXED)),
                        None)

        if not fixed_ip:
            raise q_exc.FixedIpDoesNotExistsForPort(
                fixed_ip=given_fixed_ip, port_id=port.id)

        if any(ip for ip in port.ip_addresses
               if (ip.get('address_type') in (ip_types.FLOATING,
                                              ip_types.SCALING) and
                   ip.fixed_ip['address_readable'] == given_fixed_ip)):
            raise q_exc.PortAlreadyContainsFloatingIp(
                port_id=port.id)
    return fixed_ip


def _allocate_ip(context, network, port, requested_ip_address, address_type):
    new_addresses = []
    ip_addresses = []
    if requested_ip_address:
        ip_addresses.append(requested_ip_address)

    seg_name = CONF.QUARK.floating_ip_segment_name
    strategy_name = CONF.QUARK.floating_ip_ipam_strategy

    if strategy_name.upper() == 'NETWORK':
        strategy_name = network.get("ipam_strategy")

    port_id = port
    if port:
        port_id = port.id

    ipam_driver = ipam.IPAM_REGISTRY.get_strategy(strategy_name)
    ipam_driver.allocate_ip_address(context, new_addresses, network.id,
                                    port_id, CONF.QUARK.ipam_reuse_after,
                                    seg_name, version=4,
                                    ip_addresses=ip_addresses,
                                    address_type=address_type)

    return new_addresses[0]


def _get_next_available_fixed_ip(port):
    floating_ips = [ip for ip in port.ip_addresses
                    if ip.get('address_type') in
                    (ip_types.FLOATING, ip_types.SCALING)]
    fixed_ips = [ip for ip in port.ip_addresses
                 if ip.get('address_type') == ip_types.FIXED]

    if not fixed_ips or len(fixed_ips) == 0:
        return None

    used = [ip.fixed_ip.address for ip in floating_ips
            if ip and ip.fixed_ip]

    return next((ip for ip in sorted(fixed_ips,
                                     key=lambda ip: ip.get('allocated_at'))
                if ip.address not in used), None)


def _get_ips_by_type(context, ip_type, filters=None, fields=None):
    filters = filters or {}
    filters['_deallocated'] = False
    filters['address_type'] = ip_type
    ips = db_api.floating_ip_find(context, scope=db_api.ALL, **filters)
    return ips


def _create_flip(context, flip, port_fixed_ips):
    """Associates the flip with ports and creates it with the flip driver

    :param context: neutron api request context.
    :param flip: quark.db.models.IPAddress object representing a floating IP
    :param port_fixed_ips: dictionary of the structure:
    {"<id of port>": {"port": <quark.db.models.Port>,
     "fixed_ip": "<fixed ip address>"}}
    :return: None
    """
    if port_fixed_ips:
        context.session.begin()
        try:
            ports = [val['port'] for val in port_fixed_ips.values()]
            flip = db_api.port_associate_ip(context, ports, flip,
                                            port_fixed_ips.keys())

            for port_id in port_fixed_ips:
                fixed_ip = port_fixed_ips[port_id]['fixed_ip']
                flip = db_api.floating_ip_associate_fixed_ip(context, flip,
                                                             fixed_ip)

            flip_driver = registry.DRIVER_REGISTRY.get_driver()

            flip_driver.register_floating_ip(flip, port_fixed_ips)
            context.session.commit()
        except Exception:
            context.session.rollback()
            raise

    # alexm: Notify from this method for consistency with _delete_flip
    billing.notify(context, billing.IP_ASSOC, flip)


def _get_flip_fixed_ip_by_port_id(flip, port_id):
    for fixed_ip in flip.fixed_ips:
        if fixed_ip.ports[0].id == port_id:
            return fixed_ip


def _update_flip(context, flip_id, ip_type, requested_ports):
    """Update a flip based IPAddress

    :param context: neutron api request context.
    :param flip_id: id of the flip or scip
    :param ip_type: ip_types.FLOATING | ip_types.SCALING
    :param requested_ports: dictionary of the structure:
    {"port_id": "<id of port>", "fixed_ip": "<fixed ip address>"}
    :return: quark.models.IPAddress
    """
    # This list will hold flips that require notifications.
    # Using sets to avoid dups, if any.
    notifications = {
        billing.IP_ASSOC: set(),
        billing.IP_DISASSOC: set()
    }

    context.session.begin()
    try:
        flip = db_api.floating_ip_find(context, id=flip_id, scope=db_api.ONE)
        if not flip:
            if ip_type == ip_types.SCALING:
                raise q_exc.ScalingIpNotFound(id=flip_id)
            raise q_exc.FloatingIpNotFound(id=flip_id)
        current_ports = flip.ports

        # Determine what ports are being removed, being added, and remain
        req_port_ids = [request_port.get('port_id')
                        for request_port in requested_ports]
        curr_port_ids = [curr_port.id for curr_port in current_ports]
        added_port_ids = [port_id for port_id in req_port_ids
                          if port_id and port_id not in curr_port_ids]
        removed_port_ids = [port_id for port_id in curr_port_ids
                            if port_id not in req_port_ids]
        remaining_port_ids = set(curr_port_ids) - set(removed_port_ids)

        # Validations just for floating ip types
        if (ip_type == ip_types.FLOATING and curr_port_ids and
                curr_port_ids == req_port_ids):
            d = dict(flip_id=flip_id, port_id=curr_port_ids[0])
            raise q_exc.PortAlreadyAssociatedToFloatingIp(**d)
        if (ip_type == ip_types.FLOATING and
                not curr_port_ids and not req_port_ids):
            raise q_exc.FloatingIpUpdateNoPortIdSupplied()

        # Validate that GW IP is not in use on the NW.
        flip_subnet = v._make_subnet_dict(flip.subnet)
        for added_port_id in added_port_ids:
            port = _get_port(context, added_port_id)
            nw = port.network
            nw_ports = v._make_ports_list(nw.ports)
            fixed_ips = [ip.get('ip_address') for p in nw_ports
                         for ip in p.get('fixed_ips')]

            gw_ip = flip_subnet.get('gateway_ip')
            if gw_ip in fixed_ips:
                port_with_gateway_ip = None
                for port in nw_ports:
                    for ip in port.get('fixed_ips'):
                        if gw_ip in ip.get('ip_address'):
                            port_with_gateway_ip = port
                            break
                port_id = port_with_gateway_ip.get('id')
                network_id = port_with_gateway_ip.get('network_id')
                raise q_exc.FixedIpAllocatedToGatewayIp(port_id=port_id,
                                                        network_id=network_id)
        port_fixed_ips = {}

        # Keep the ports and fixed ips that have not changed
        for port_id in remaining_port_ids:
            port = db_api.port_find(context, id=port_id, scope=db_api.ONE)
            fixed_ip = _get_flip_fixed_ip_by_port_id(flip, port_id)
            port_fixed_ips[port_id] = {'port': port, 'fixed_ip': fixed_ip}

        # Disassociate the ports and fixed ips from the flip that were
        # associated to the flip but are not anymore
        for port_id in removed_port_ids:
            port = db_api.port_find(context, id=port_id, scope=db_api.ONE)
            flip = db_api.port_disassociate_ip(context, [port], flip)
            notifications[billing.IP_DISASSOC].add(flip)
            fixed_ip = _get_flip_fixed_ip_by_port_id(flip, port_id)
            if fixed_ip:
                flip = db_api.floating_ip_disassociate_fixed_ip(
                    context, flip, fixed_ip)

        # Validate the new ports with the flip and associate the new ports
        # and fixed ips with the flip
        for port_id in added_port_ids:
            port = db_api.port_find(context, id=port_id, scope=db_api.ONE)
            if not port:
                raise n_exc.PortNotFound(port_id=port_id)
            if any(ip for ip in port.ip_addresses
                   if (ip.get('address_type') == ip_types.FLOATING)):
                raise q_exc.PortAlreadyContainsFloatingIp(port_id=port_id)
            if any(ip for ip in port.ip_addresses
                   if (ip.get('address_type') == ip_types.SCALING)):
                raise q_exc.PortAlreadyContainsScalingIp(port_id=port_id)
            fixed_ip = _get_next_available_fixed_ip(port)
            LOG.info('new fixed ip: %s' % fixed_ip)
            if not fixed_ip:
                raise q_exc.NoAvailableFixedIpsForPort(port_id=port_id)
            port_fixed_ips[port_id] = {'port': port, 'fixed_ip': fixed_ip}
            flip = db_api.port_associate_ip(context, [port], flip, [port_id])
            notifications[billing.IP_ASSOC].add(flip)
            flip = db_api.floating_ip_associate_fixed_ip(context, flip,
                                                         fixed_ip)

        flip_driver = registry.DRIVER_REGISTRY.get_driver()
        # If there are not any remaining ports and no new ones are being added,
        # remove the floating ip from unicorn
        if not remaining_port_ids and not added_port_ids:
            flip_driver.remove_floating_ip(flip)
        # If new ports are being added but there previously was not any ports,
        # then register a new floating ip with the driver because it is
        # assumed it does not exist
        elif added_port_ids and not curr_port_ids:
            flip_driver.register_floating_ip(flip, port_fixed_ips)
        else:
            flip_driver.update_floating_ip(flip, port_fixed_ips)
        context.session.commit()
    except Exception:
        context.session.rollback()
        raise

    # Send notifications for possible associate/disassociate events
    for notif_type, flip_set in notifications.iteritems():
        for flip in flip_set:
            billing.notify(context, notif_type, flip)

    # NOTE(blogan): ORM does not seem to update the model to the real state
    # of the database, so I'm doing an explicit refresh for now.
    context.session.refresh(flip)
    return flip


def _delete_flip(context, id, address_type):
    filters = {'address_type': address_type, '_deallocated': False}

    flip = db_api.floating_ip_find(context, id=id, scope=db_api.ONE, **filters)
    if not flip:
        raise q_exc.FloatingIpNotFound(id=id)

    current_ports = flip.ports
    if address_type == ip_types.FLOATING:
        if current_ports:
            current_ports = [flip.ports[0]]
    elif address_type == ip_types.SCALING:
        current_ports = flip.ports

    context.session.begin()
    try:
        strategy_name = flip.network.get('ipam_strategy')
        ipam_driver = ipam.IPAM_REGISTRY.get_strategy(strategy_name)
        ipam_driver.deallocate_ip_address(context, flip)

        if current_ports:
            db_api.port_disassociate_ip(context, current_ports, flip)
        if flip.fixed_ips:
            db_api.floating_ip_disassociate_all_fixed_ips(context, flip)

        context.session.commit()
    except Exception:
        context.session.rollback()
        raise

    try:
        driver = registry.DRIVER_REGISTRY.get_driver()
        driver.remove_floating_ip(flip)
    except Exception as e:
        LOG.error('There was an error when trying to delete the floating ip '
                  'on the unicorn API.  The ip has been cleaned up, but '
                  'may need to be handled manually in the unicorn API.  '
                  'Error: %s' % e.message)

    # alexm: Notify from this method because we don't have the flip object
    # in the callers
    billing.notify(context, billing.IP_DISASSOC, flip)


def create_floatingip(context, content):
    """Allocate or reallocate a floating IP.

    :param context: neutron api request context.
    :param content: dictionary describing the floating ip, with keys
        as listed in the RESOURCE_ATTRIBUTE_MAP object in
        neutron/api/v2/attributes.py.  All keys will be populated.

    :returns: Dictionary containing details for the new floating IP.  If values
        are declared in the fields parameter, then only those keys will be
        present.
    """
    LOG.info('create_floatingip %s for tenant %s and body %s' %
             (id, context.tenant_id, content))
    network_id = content.get('floating_network_id')
    # TODO(blogan): Since the extension logic will reject any requests without
    # floating_network_id, is this still needed?
    if not network_id:
        raise n_exc.BadRequest(resource='floating_ip',
                               msg='floating_network_id is required.')
    fixed_ip_address = content.get('fixed_ip_address')
    ip_address = content.get('floating_ip_address')
    port_id = content.get('port_id')
    port = None
    port_fixed_ip = {}

    network = _get_network(context, network_id)
    if port_id:
        port = _get_port(context, port_id)
        fixed_ip = _get_fixed_ip(context, fixed_ip_address, port)
        port_fixed_ip = {port.id: {'port': port, 'fixed_ip': fixed_ip}}
    flip = _allocate_ip(context, network, port, ip_address, ip_types.FLOATING)
    _create_flip(context, flip, port_fixed_ip)
    return v._make_floating_ip_dict(flip, port_id)


def update_floatingip(context, id, content):
    """Update an existing floating IP.

    :param context: neutron api request context.
    :param id: id of the floating ip
    :param content: dictionary with keys indicating fields to update.
        valid keys are those that have a value of True for 'allow_put'
        as listed in the RESOURCE_ATTRIBUTE_MAP object in
        neutron/api/v2/attributes.py.

    :returns: Dictionary containing details for the new floating IP.  If values
        are declared in the fields parameter, then only those keys will be
        present.
    """

    LOG.info('update_floatingip %s for tenant %s and body %s' %
             (id, context.tenant_id, content))

    if 'port_id' not in content:
        raise n_exc.BadRequest(resource='floating_ip',
                               msg='port_id is required.')

    requested_ports = []
    if content.get('port_id'):
        requested_ports = [{'port_id': content.get('port_id')}]
    flip = _update_flip(context, id, ip_types.FLOATING, requested_ports)
    return v._make_floating_ip_dict(flip)


def delete_floatingip(context, id):
    """deallocate a floating IP.

    :param context: neutron api request context.
    :param id: id of the floating ip
    """

    LOG.info('delete_floatingip %s for tenant %s' % (id, context.tenant_id))

    _delete_flip(context, id, ip_types.FLOATING)


def get_floatingip(context, id, fields=None):
    """Retrieve a floating IP.

    :param context: neutron api request context.
    :param id: The UUID of the floating IP.
    :param fields: a list of strings that are valid keys in a
        floating IP dictionary as listed in the RESOURCE_ATTRIBUTE_MAP
        object in neutron/api/v2/attributes.py. Only these fields
        will be returned.

    :returns: Dictionary containing details for the floating IP.  If values
        are declared in the fields parameter, then only those keys will be
        present.
    """
    LOG.info('get_floatingip %s for tenant %s' % (id, context.tenant_id))

    filters = {'address_type': ip_types.FLOATING, '_deallocated': False}

    floating_ip = db_api.floating_ip_find(context, id=id, scope=db_api.ONE,
                                          **filters)

    if not floating_ip:
        raise q_exc.FloatingIpNotFound(id=id)

    return v._make_floating_ip_dict(floating_ip)


def get_floatingips(context, filters=None, fields=None, sorts=None, limit=None,
                    marker=None, page_reverse=False):
    """Retrieve a list of floating ips.

    :param context: neutron api request context.
    :param filters: a dictionary with keys that are valid keys for
        a floating ip as listed in the RESOURCE_ATTRIBUTE_MAP object
        in neutron/api/v2/attributes.py.  Values in this dictionary
        are an iterable containing values that will be used for an exact
        match comparison for that value.  Each result returned by this
        function will have matched one of the values for each key in
        filters.
    :param fields: a list of strings that are valid keys in a
        floating IP dictionary as listed in the RESOURCE_ATTRIBUTE_MAP
        object in neutron/api/v2/attributes.py. Only these fields
        will be returned.

    :returns: List of floating IPs that are accessible to the tenant who
        submits the request (as indicated by the tenant id of the context)
        as well as any filters.
    """
    LOG.info('get_floatingips for tenant %s filters %s fields %s' %
             (context.tenant_id, filters, fields))

    floating_ips = _get_ips_by_type(context, ip_types.FLOATING,
                                    filters=filters, fields=fields)

    return [v._make_floating_ip_dict(flip) for flip in floating_ips]


def get_floatingips_count(context, filters=None):
    """Return the number of floating IPs.

    :param context: neutron api request context
    :param filters: a dictionary with keys that are valid keys for
        a floating IP as listed in the RESOURCE_ATTRIBUTE_MAP object
        in neutron/api/v2/attributes.py.  Values in this dictionary
        are an iterable containing values that will be used for an exact
        match comparison for that value.  Each result returned by this
        function will have matched one of the values for each key in
        filters.

    :returns: The number of floating IPs that are accessible to the tenant who
        submits the request (as indicated by the tenant id of the context)
        as well as any filters.

    NOTE: this method is optional, as it was not part of the originally
          defined plugin API.
    """
    LOG.info('get_floatingips_count for tenant %s filters %s' %
             (context.tenant_id, filters))

    if filters is None:
        filters = {}

    filters['_deallocated'] = False
    filters['address_type'] = ip_types.FLOATING
    count = db_api.ip_address_count_all(context, filters)

    LOG.info('Found %s floating ips for tenant %s' % (count,
                                                      context.tenant_id))
    return count


def create_scalingip(context, content):
    """Allocate or reallocate a scaling IP.

    :param context: neutron api request context.
    :param content: dictionary describing the scaling ip, with keys
        as listed in the RESOURCE_ATTRIBUTE_MAP object in
        neutron/api/v2/attributes.py.  All keys will be populated.

    :returns: Dictionary containing details for the new scaling IP.  If values
        are declared in the fields parameter, then only those keys will be
        present.
    """
    LOG.info('create_scalingip for tenant %s and body %s',
             context.tenant_id, content)
    network_id = content.get('scaling_network_id')
    ip_address = content.get('scaling_ip_address')
    requested_ports = content.get('ports', [])

    network = _get_network(context, network_id)
    port_fixed_ips = {}
    for req_port in requested_ports:
        port = _get_port(context, req_port['port_id'])
        fixed_ip = _get_fixed_ip(context, req_port.get('fixed_ip_address'),
                                 port)
        port_fixed_ips[port.id] = {"port": port, "fixed_ip": fixed_ip}
    scip = _allocate_ip(context, network, None, ip_address, ip_types.SCALING)
    _create_flip(context, scip, port_fixed_ips)
    return v._make_scaling_ip_dict(scip)


def update_scalingip(context, id, content):
    """Update an existing scaling IP.

    :param context: neutron api request context.
    :param id: id of the scaling ip
    :param content: dictionary with keys indicating fields to update.
        valid keys are those that have a value of True for 'allow_put'
        as listed in the RESOURCE_ATTRIBUTE_MAP object in
        neutron/api/v2/attributes.py.

    :returns: Dictionary containing details for the new scaling IP.  If values
        are declared in the fields parameter, then only those keys will be
        present.
    """
    LOG.info('update_scalingip %s for tenant %s and body %s' %
             (id, context.tenant_id, content))
    requested_ports = content.get('ports', [])
    flip = _update_flip(context, id, ip_types.SCALING, requested_ports)
    return v._make_scaling_ip_dict(flip)


def delete_scalingip(context, id):
    """Deallocate a scaling IP.

    :param context: neutron api request context.
    :param id: id of the scaling ip
    """
    LOG.info('delete_scalingip %s for tenant %s' % (id, context.tenant_id))
    _delete_flip(context, id, ip_types.SCALING)


def get_scalingip(context, id, fields=None):
    """Retrieve a scaling IP.

    :param context: neutron api request context.
    :param id: The UUID of the scaling IP.
    :param fields: a list of strings that are valid keys in a
        scaling IP dictionary as listed in the RESOURCE_ATTRIBUTE_MAP
        object in neutron/api/v2/attributes.py. Only these fields
        will be returned.

    :returns: Dictionary containing details for the scaling IP.  If values
        are declared in the fields parameter, then only those keys will be
        present.
    """
    LOG.info('get_scalingip %s for tenant %s' % (id, context.tenant_id))
    filters = {'address_type': ip_types.SCALING, '_deallocated': False}
    scaling_ip = db_api.floating_ip_find(context, id=id, scope=db_api.ONE,
                                         **filters)
    if not scaling_ip:
        raise q_exc.ScalingIpNotFound(id=id)
    return v._make_scaling_ip_dict(scaling_ip)


def get_scalingips(context, filters=None, fields=None, sorts=None, limit=None,
                   marker=None, page_reverse=False):
    """Retrieve a list of scaling ips.

    :param context: neutron api request context.
    :param filters: a dictionary with keys that are valid keys for
        a scaling ip as listed in the RESOURCE_ATTRIBUTE_MAP object
        in neutron/api/v2/attributes.py.  Values in this dictionary
        are an iterable containing values that will be used for an exact
        match comparison for that value.  Each result returned by this
        function will have matched one of the values for each key in
        filters.
    :param fields: a list of strings that are valid keys in a
        scaling IP dictionary as listed in the RESOURCE_ATTRIBUTE_MAP
        object in neutron/api/v2/attributes.py. Only these fields
        will be returned.

    :returns: List of scaling IPs that are accessible to the tenant who
        submits the request (as indicated by the tenant id of the context)
        as well as any filters.
    """
    LOG.info('get_scalingips for tenant %s filters %s fields %s' %
             (context.tenant_id, filters, fields))
    scaling_ips = _get_ips_by_type(context, ip_types.SCALING,
                                   filters=filters, fields=fields)
    return [v._make_scaling_ip_dict(scip) for scip in scaling_ips]



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from neutron_lib import exceptions as n_exc
from oslo_log import log as logging

from quark.db import api as db_api
from quark import exceptions as q_exc
from quark import plugin_views as v

LOG = logging.getLogger(__name__)


def _to_mac_range(val):
    cidr_parts = val.split("/")
    prefix = cidr_parts[0]

    # FIXME(anyone): replace is slow, but this doesn't really
    #               get called ever. Fix maybe?
    prefix = prefix.replace(':', '')
    prefix = prefix.replace('-', '')
    prefix_length = len(prefix)
    if prefix_length < 6 or prefix_length > 12:
        raise q_exc.InvalidMacAddressRange(cidr=val)

    diff = 12 - len(prefix)
    if len(cidr_parts) > 1:
        mask = int(cidr_parts[1])
    else:
        mask = 48 - diff * 4
    mask_size = 1 << (48 - mask)
    prefix = "%s%s" % (prefix, "0" * diff)
    try:
        cidr = "%s/%s" % (str(netaddr.EUI(prefix)).replace("-", ":"), mask)
    except netaddr.AddrFormatError:
        raise q_exc.InvalidMacAddressRange(cidr=val)
    prefix_int = int(prefix, base=16)
    return cidr, prefix_int, prefix_int + mask_size


def get_mac_address_range(context, id, fields=None):
    """Retrieve a mac_address_range.

    : param context: neutron api request context
    : param id: UUID representing the network to fetch.
    : param fields: a list of strings that are valid keys in a
        network dictionary as listed in the RESOURCE_ATTRIBUTE_MAP
        object in neutron/api/v2/attributes.py. Only these fields
        will be returned.
    """
    LOG.info("get_mac_address_range %s for tenant %s fields %s" %
             (id, context.tenant_id, fields))

    if not context.is_admin:
        raise n_exc.NotAuthorized()

    mac_address_range = db_api.mac_address_range_find(
        context, id=id, scope=db_api.ONE)

    if not mac_address_range:
        raise q_exc.MacAddressRangeNotFound(
            mac_address_range_id=id)
    return v._make_mac_range_dict(mac_address_range)


def get_mac_address_ranges(context):
    LOG.info("get_mac_address_ranges for tenant %s" % context.tenant_id)
    if not context.is_admin:
        raise n_exc.NotAuthorized()

    ranges = db_api.mac_address_range_find(context)
    return [v._make_mac_range_dict(m) for m in ranges]


def create_mac_address_range(context, mac_range):
    LOG.info("create_mac_address_range for tenant %s" % context.tenant_id)
    if not context.is_admin:
        raise n_exc.NotAuthorized()

    if not mac_range or "mac_address_range" not in mac_range:
        raise n_exc.BadRequest(resource="mac_address_range",
                               msg="Malformed body")

    rng = mac_range.get("mac_address_range")
    cidr = rng.get("cidr")

    if not cidr:
        raise n_exc.BadRequest(resource="mac_address_range",
                               msg="Missing cidr")
    do_not_use = rng.get("do_not_use", "0")
    cidr, first_address, last_address = _to_mac_range(cidr)
    with context.session.begin():
        new_range = db_api.mac_address_range_create(
            context, cidr=cidr, first_address=first_address,
            last_address=last_address, next_auto_assign_mac=first_address,
            do_not_use=do_not_use)
    return v._make_mac_range_dict(new_range)


def _delete_mac_address_range(context, mac_address_range):
    if mac_address_range.allocated_macs:
        raise q_exc.MacAddressRangeInUse(
            mac_address_range_id=mac_address_range["id"])
    db_api.mac_address_range_delete(context, mac_address_range)


def delete_mac_address_range(context, id):
    """Delete a mac_address_range.

    : param context: neutron api request context
    : param id: UUID representing the mac_address_range to delete.
    """
    LOG.info("delete_mac_address_range %s for tenant %s" %
             (id, context.tenant_id))
    if not context.is_admin:
        raise n_exc.NotAuthorized()

    with context.session.begin():
        mar = db_api.mac_address_range_find(context, id=id, scope=db_api.ONE)
        if not mar:
            raise q_exc.MacAddressRangeNotFound(
                mac_address_range_id=id)
        _delete_mac_address_range(context, mar)



import netaddr

from quark.db import api as db_api
from quark.tests.functional.mysql.base import MySqlBaseFunctionalTest

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class QuarkLocks(MySqlBaseFunctionalTest):
    def test_create_lock_holder(self):
        kwargs = {"address": netaddr.IPAddress("192.168.2.1")}
        ip_address = db_api.ip_address_create(self.context, **kwargs)
        kwargs = {"type": "ip_address", "name": "because i said so"}
        lock_holder = db_api.lock_holder_create(
            self.context, ip_address, **kwargs)

        self.context.session.refresh(ip_address)
        self.assertEqual(ip_address.lock_id, lock_holder.lock_id)

    def test_create_lock_holder_already_locked(self):
        kwargs = {"address": netaddr.IPAddress("192.168.2.1")}
        ip_address = db_api.ip_address_create(self.context, **kwargs)
        name = "because i said so"
        kwargs_1 = {"type": "ip_address", "name": name + "1"}
        lock_holder_1 = db_api.lock_holder_create(
            self.context, ip_address, **kwargs_1)
        self.context.session.flush()

        kwargs_2 = {"type": "ip_address", "name": name + "2"}
        self.context.session.refresh(ip_address)
        lock_holder_2 = db_api.lock_holder_create(
            self.context, ip_address, **kwargs_2)
        self.context.session.flush()

        self.context.session.refresh(ip_address)
        self.assertNotEqual(lock_holder_1.id, lock_holder_2.id)
        self.assertEqual(lock_holder_1.name, name + "1")
        self.assertEqual(lock_holder_2.name, name + "2")
        self.assertEqual(lock_holder_1.lock_id, lock_holder_2.lock_id)
        self.assertEqual(ip_address.lock_id, lock_holder_1.lock_id)

    def test_find_lock_holder(self):
        kwargs = {"address": netaddr.IPAddress("192.168.2.1")}
        ip_address = db_api.ip_address_create(self.context, **kwargs)
        kwargs = {"type": "ip_address", "name": "because i said so"}
        lock_holder = db_api.lock_holder_create(
            self.context, ip_address, **kwargs)

        self.context.session.refresh(ip_address)
        self.assertEqual(ip_address.lock_id, lock_holder.lock_id)

        lock_holders = db_api.lock_holder_find(
            self.context,
            lock_id=ip_address.lock_id, name=kwargs["name"],
            scope=db_api.ALL)
        self.assertEqual(len(lock_holders), 1)
        self.assertEqual(lock_holders[0]["lock_id"], ip_address.lock_id)
        self.assertEqual(lock_holders[0]["name"], kwargs["name"])

    def test_delete_lock_holder(self):
        kwargs = {"address": netaddr.IPAddress("192.168.2.1")}
        ip_address = db_api.ip_address_create(self.context, **kwargs)
        kwargs = {"type": "ip_address", "name": "because i said so"}
        lock_holder = db_api.lock_holder_create(
            self.context, ip_address, **kwargs)
        self.context.session.flush()
        self.context.session.refresh(ip_address)
        self.assertEqual(ip_address.lock_id, lock_holder.lock_id)

        db_api.lock_holder_delete(self.context, ip_address, lock_holder)
        self.context.session.refresh(ip_address)
        self.assertIsNone(ip_address.lock_id)



"""add_type_column_quark_ip_addresses

Revision ID: 42a3c8c0db75
Revises: 3b467be51e43
Create Date: 2014-10-14 14:21:25.016371

"""

# revision identifiers, used by Alembic.
revision = '42a3c8c0db75'
down_revision = '3b467be51e43'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('quark_ip_addresses',
                  sa.Column('address_type',
                            sa.Enum('shared', 'floating', 'fixed',
                                    name='quark_ip_address_types'),
                            nullable=True))


def downgrade():
    op.drop_column('quark_ip_addresses', 'address_type')



"""Added Floating IP to Fixed IP mapping table

Revision ID: 33e9e23ba761
Revises: 356d6c0623c8
Create Date: 2015-05-11 14:14:23.619952

"""

# revision identifiers, used by Alembic.
revision = '33e9e23ba761'
down_revision = '356d6c0623c8'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('quark_floating_to_fixed_ip_address_associations',
                    sa.Column('floating_ip_address_id', sa.String(length=36),
                              nullable=False),
                    sa.Column('fixed_ip_address_id', sa.String(length=36),
                              nullable=False),
                    sa.Column('enabled', sa.Boolean(), server_default='1',
                              nullable=False),
                    sa.ForeignKeyConstraint(['fixed_ip_address_id'],
                                            ['quark_ip_addresses.id'], ),
                    sa.ForeignKeyConstraint(['floating_ip_address_id'],
                                            ['quark_ip_addresses.id'], ),
                    sa.PrimaryKeyConstraint('floating_ip_address_id',
                                            'fixed_ip_address_id'),
                    mysql_engine='InnoDB')


def downgrade():
    op.drop_table('quark_floating_to_fixed_ip_address_associations')
